#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
`nigsp` main workflow and related functions.

The workflow is callable either as a python function, or (preferably)
in a shell session:
```shell
$ nigsp --help
```
"""
import datetime
import logging
import os
import sys

import numpy as np

from nigsp import _version, blocks, io, references
from nigsp import surrogates as surr
from nigsp import timeseries as ts
from nigsp import utils, viz
from nigsp.cli.run import _get_parser
from nigsp.due import due
from nigsp.objects import SCGraph
from nigsp.operations.metrics import SUPPORTED_METRICS

LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def save_bash_call(fname, outdir, outname):
    """
    Save the bash call into file `p2d_call.sh`.

    Parameters
    ----------
    outdir : str or os.PathLike, optional
        output directory
    """
    fname = utils.change_var_type(fname, list, stop=False, silent=True)

    # Prepare folders
    if outdir is None:
        if outname is None:
            if len(fname) == 1:
                outdir = os.path.dirname(fname[0])
            else:
                outdir = os.path.commonpath(fname)
        else:
            outdir = os.path.split(outname)[0]

        if outdir == "" or outdir == "/":
            outdir = "."
        outdir = os.path.join(outdir, "nigsp")

    outdir = os.path.abspath(outdir)
    log_path = os.path.join(outdir, "logs")
    os.makedirs(log_path, exist_ok=True)
    arg_str = " ".join(sys.argv[1:])
    call_str = f"nigsp {arg_str}"
    outdir = os.path.abspath(outdir)
    log_path = os.path.join(outdir, "logs")
    os.makedirs(log_path, exist_ok=True)
    isotime = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
    f = open(os.path.join(log_path, f"nigsp_call_{isotime}.sh"), "a")
    f.write(f"#!bin/bash \n{call_str}")
    f.close()


@due.dcite(references.PRETI_2019)
@due.dcite(references.NIGSP)
def nigsp(
    fname,
    scname,
    atlasname=None,
    outname=None,
    outdir=None,
    comp_metric=[],
    index="median",
    surr_type=None,
    n_surr=1000,
    method="Bernoulli",
    p=0.05,
    seed=None,
    lgr_degree="info",
):
    """
    Main workflow for nigsp, following the methods described in [1].

    Parameters
    ----------
    fname : str or os.PathLike
        Path to the timeseries data file. It can be a text, nifti, or matlab file
        (and variants). To see the full list of support, check the general documentation.
    scname : str or os.PathLike
        Path to the structural connectivity data file. It can be a text, nifti, or matlab file
        (and variants). To see the full list of support, check the general documentation.
    atlasname : str, os.PathLike, or None, optional
        Path to the atlas data file. It can be a text, nifti, or matlab file
        (and variants). To see the full list of support, check the general documentation.
    outname : str, os.PathLike, or None, optional
        Path to the output file - or just its full name. It can be a text, nifti, or matlab file
        (and variants). If an extension is *not* declared, or if it is not currently
        supported, the program will automatically export a csv file.
        To see the full list of support, check the general documentation.
        It is *not* necessary to declare both this and `outdir` - the full path can
        be specified here.
    outdir : str, os.PathLike, or None, optional
        Path to the output folder. If it doesn't exist, it will be created.
        If both `outdir` and `outname` are declared, `outdir` overrides the path
        specified in `outname` (but not the filename!)
    comp_metric : list or None, optional
        List of metrics that should be computed. If empty (default), compute all
        metrics available.
    index : 'median' or int, optional
        The index of the eigenvector/harmonic of the graph to split the graph in
        multiple parts, or the method to find this index. Currently supports
        median energy split ('median'), which is the default.
        Note that indexing is one-based (i.e. indexing starts at 1, not 0).
    surr_type : 'informed', 'uninformed', or None, optional
        The type of surrogates to create for statistical testing.
        'Informed' surrogates are created using the empirical SC decomposition,
        shuffling eigenvectors' signs. 'Uninformed' surrogates use a configuration
        model of the same degree of the empirical SC instead.
        If surr_type is None, no statistical test is run.
    n_surr : int, optional
        Number of surrogates to be created.
    method : 'Bernoulli', 'frequentist', or None, optional
        Method to use for statistical testing or empirical data vs surrogates.
        Supported possibilities are 'Bernoulli', to test empirical data at
        the group level against a random Bernoulli process,
        or 'frequentist', to test empirical data at the group or subject level
        against the plain surrogate distribution.
    p : float (in range [0 1]), optional
        The two-tailed p value to use for testing. Must be between 0 and 1.
        The two tails will be tested against half of this value.
    seed : int or None, optional
        The seed to use to reinitialise the random number generator for
        surrogates creation.
        If `seed` is None, it will use internally specified seeds to guarantee
        replicability. It's suggested either not specify it, or specify a different
        one for each type of surrogate created.
    lgr_degree : 'debug', 'info', or 'quiet', optional
        The degree of verbosity of the logger. Default is 'info'.

    Returns
    -------
    0
        If there are no errors.

    Raises
    ------
    NotImplementedError
        If SC file is not of a supported type.
        If timeseries file is not of a supported type.
        If atlas file is not of a supported type.
        If timeseries file is a nifti file but atlas file is not.
        If statistical method is not within supported method (or None).
        If surrogate type is not within supported type (or None).
    ValueError
        If `index` is not int or is not `median`.
        If `p` is not in the range [0 1].
        If the projected timeseries are not split to compute SDI or gSDI.

    See also
    --------
    [1] Preti, M.G., Van De Ville, D. Decoupling of brain function from structure
    reveals regional behavioral specialization in humans. Nat Commun 10, 4747 (2019).
    https://doi.org/10.1038/s41467-019-12765-7

    Configuration model: https://en.wikipedia.org/wiki/Configuration_model
    """
    # #### Logger preparation #### #
    fname = utils.change_var_type(fname, list, stop=False, silent=True)

    # Prepare folders
    if outdir is None:
        if outname is None:
            if len(fname) == 1:
                outdir = os.path.dirname(fname[0])
            else:
                outdir = os.path.commonpath(fname)
        else:
            outdir = os.path.split(outname)[0]

        if outdir == "" or outdir == "/":
            outdir = "."
        outdir = os.path.join(outdir, "nigsp")

    outdir = os.path.abspath(outdir)
    log_path = os.path.join(outdir, "logs")
    os.makedirs(log_path, exist_ok=True)

    # Create logfile name
    basename = "nigsp_"
    extension = "tsv"
    isotime = datetime.datetime.now().strftime("%Y-%m-%dT%H%M%S")
    logname = os.path.join(log_path, f"{basename}{isotime}.{extension}")

    # Set logging format
    log_formatter = logging.Formatter(
        "%(asctime)s\t%(name)-12s\t%(levelname)-8s\t%(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S",
    )

    # Set up logging file and open it for writing
    log_handler = logging.FileHandler(logname)
    log_handler.setFormatter(log_formatter)
    sh = logging.StreamHandler()

    if lgr_degree == "quiet":
        logging.basicConfig(
            level=logging.WARNING,
            handlers=[log_handler, sh],
            format="%(levelname)-10s %(message)s",
        )
    elif lgr_degree == "debug":
        logging.basicConfig(
            level=logging.DEBUG,
            handlers=[log_handler, sh],
            format="%(levelname)-10s %(message)s",
        )
    else:
        logging.basicConfig(
            level=logging.INFO,
            handlers=[log_handler, sh],
            format="%(levelname)-10s %(message)s",
        )

    version_number = _version.get_versions()["version"]
    LGR.info(f"Currently running nigsp version {version_number}")

    # #### Check input #### #

    # Check data files
    LGR.info(f"Input structural connectivity file: {scname}")
    sc_is = dict.fromkeys(io.EXT_DICT.keys(), False)
    LGR.info(f"Input functional file(s): {fname}")
    func_is = dict.fromkeys(io.EXT_DICT.keys(), "")
    atlas_is = dict.fromkeys(io.EXT_DICT.keys(), False)
    if atlasname:
        LGR.info(f"Input atlas file: {atlasname}")

    # Check inputs type
    for k in io.EXT_DICT.keys():
        func_is[k] = []
        for f in fname:
            func_is[k] += [io.check_ext(io.EXT_DICT[k], f)[0]]
        # Check that func files are all of the same kind
        func_is[k] = all(func_is[k])

        sc_is[k], _ = io.check_ext(io.EXT_DICT[k], scname)

        if atlasname is not None:
            atlas_is[k], _ = io.check_ext(io.EXT_DICT[k], atlasname)

    # Check that other inputs are supported
    if index != "median" and type(index) is not int:
        raise ValueError(f"Index {index} of type {type(index)} is not valid.")
    if method not in surr.STAT_METHOD and method is not None:
        raise NotImplementedError(
            f"Method {method} is not supported. Supported "
            f"methods are: {surr.STAT_METHOD}"
        )
    if surr_type not in surr.SURR_TYPE and surr_type is not None:
        raise NotImplementedError(
            f"Surrogate type {surr_type} is not supported. "
            f"Supported types are: {surr.SURR_TYPE}"
        )
    if p < 0 or p > 1:
        raise ValueError(
            "P value must be between 0 and 1, but {p} was provided instead."
        )

    # Check what metric to compute
    if comp_metric not in [[], None]:
        for item in comp_metric:
            if item not in SUPPORTED_METRICS:
                raise NotImplementedError(
                    f"Metric {item} is not supported. "
                    f"Supported metrics are: {SUPPORTED_METRICS}"
                )
    else:
        comp_metric = SUPPORTED_METRICS

    # #### Prepare Outputs #### #
    if outname is not None:
        _, outprefix, outext = io.check_ext(io.EXT_ALL, outname, remove=True)
        outprefix = os.path.join(outdir, f"{os.path.split(outprefix)[1]}_")
    else:
        outprefix = f"{outdir}{os.sep}"

    # #### Read in data #### #

    # Read in structural connectivity matrix
    if sc_is["1D"]:
        mtx = io.load_txt(scname, shape="square")
    elif sc_is["mat"]:
        mtx = io.load_mat(scname, shape="square")
    elif sc_is["xls"]:
        mtx = io.load_xls(scname, shape="square")
    else:
        raise NotImplementedError(f"Input file {scname} is not of a supported type.")

    # Read in atlas, if defined
    if atlasname is not None:
        if (
            atlas_is["1D"] or atlas_is["mat"] or atlas_is["xls"] or atlas_is["nifti"]
        ) is False:
            raise NotImplementedError(
                f"Input file {atlasname} is not of a supported type."
            )
        elif atlas_is["1D"]:
            atlas = io.load_txt(atlasname)
        elif atlas_is["nifti"]:
            atlas, _, img = io.load_nifti_get_mask(atlasname, ndim=3)
        elif atlas_is["mat"]:
            atlas = io.load_mat(atlasname)
        elif atlas_is["xls"]:
            atlas = io.load_xls(atlasname)
    else:
        LGR.warning("Atlas not provided. Some functionalities might not work.")
        atlas, img = None, None

    # Read in functional timeseries, join them, and normalise them
    timeseries = []
    for f in fname:
        if func_is["nifti"] and atlas_is["nifti"]:
            t, atlas, img = blocks.nifti_to_timeseries(f, atlasname)
        elif func_is["nifti"] and atlas_is["nifti"] is False:
            raise NotImplementedError(
                "To work with functional file(s) of nifti format, "
                "specify an atlas file in nifti format."
            )
        elif func_is["1D"]:
            t = io.load_txt(f)
        elif func_is["mat"]:
            t = io.load_mat(f)
        elif func_is["xls"]:
            t = io.load_xls(f)
        else:
            raise NotImplementedError(f"Input file {f} is not of a supported type.")

        timeseries += [t[..., np.newaxis]]

    timeseries = np.concatenate(timeseries, axis=-1).squeeze()
    timeseries = ts.normalise_ts(timeseries)

    # #### Assign SCGraph object #### #
    scgraph = SCGraph(mtx, timeseries, atlas=atlas, img=img)

    # #### Compute SDI (split low vs high timeseries) and FC and output them #### #

    # Run laplacian decomposition and actually filter timeseries.
    LGR.info("Run laplacian decomposition of structural graph.")
    scgraph.structural_decomposition()
    LGR.info("Compute the energy of the graph and split it in parts.")
    scgraph.compute_graph_energy(mean=True).split_graph()

    if "sdi" in comp_metric or "gsdi" in comp_metric:
        # If there are more than two splits in the timeseries, compute Generalised SDI
        # This should not happen in this moment.
        if len(scgraph.split_keys) == 2:
            metric_name = "sdi"
            scgraph.compute_sdi()
        elif len(scgraph.split_keys) > 2:
            metric_name = "gsdi"
            scgraph.compute_gsdi()
        # Export non-thresholded metrics
        LGR.info(f"Export non-thresholded version of {metric_name}.")
        blocks.export_metric(scgraph, outext, outprefix)

    if "dfc" in comp_metric or "fc" in comp_metric:
        scgraph.compute_fc(mean=True)
        for k in scgraph.split_keys:
            LGR.info(f"Export {k} FC (data).")
            io.export_mtx(scgraph.fc_split[k], f"{outprefix}fc_{k}", ext=outext)
        # Export fc
        LGR.info("Export original FC (data).")
        io.export_mtx(scgraph.fc, f"{outprefix}fc", ext=outext)

    # #### Output more results (pt. 1) #### #

    # Export eigenvalues, eigenvectors, and split timeseries and eigenvectors
    for k in scgraph.split_keys:
        LGR.info(f"Export {k} timeseries.")
        io.export_mtx(scgraph.ts_split[k], f"{outprefix}timeseries_{k}", ext=outext)
        LGR.info(f"Export {k} eigenvectors.")
        io.export_mtx(scgraph.evec_split[k], f"{outprefix}eigenvec_{k}", ext=outext)
    LGR.info("Export original eigenvectors.")
    io.export_mtx(scgraph.eigenvec, f"{outprefix}eigenvec", ext=outext)
    LGR.info("Export original eigenvalues.")
    io.export_mtx(scgraph.eigenval, f"{outprefix}eigenval", ext=outext)

    # #### Additional steps #### #

    # If possible, create plots!
    try:
        import matplotlib as _
        import nilearn as _

        # Plot original SC and Laplacian
        LGR.info("Plot laplacian matrix.")
        viz.plot_connectivity(scgraph.lapl_mtx, f"{outprefix}laplacian.png")
        LGR.info("Plot structural connectivity matrix.")
        viz.plot_connectivity(scgraph.mtx, f"{outprefix}sc.png")

        # Plot timeseries
        LGR.info("Plot original timeseries.")
        viz.plot_greyplot(scgraph.timeseries, f"{outprefix}greyplot.png")
        for k in scgraph.split_keys:
            LGR.info(f"Plot {k} timeseries.")
            viz.plot_greyplot(scgraph.ts_split[k], f"{outprefix}greyplot_{k}.png")

        if "dfc" in comp_metric or "fc" in comp_metric:
            # Plot FC
            LGR.info("Plot original functional connectivity matrix.")
            viz.plot_connectivity(scgraph.fc, f"{outprefix}fc.png")
            for k in scgraph.split_keys:
                LGR.info(f"Plot {k} functional connectivity matrix.")
                viz.plot_connectivity(scgraph.fc_split[k], f"{outprefix}fc_{k}.png")
        if "sdi" in comp_metric or "gsdi" in comp_metric:
            if atlasname is not None:
                LGR.info(f"Plot {metric_name} markerplot.")
                if img is not None:
                    blocks.plot_metric(scgraph, outprefix, img)
                elif atlas is not None:
                    blocks.plot_metric(scgraph, outprefix, atlas)

    except ImportError:
        LGR.warning(
            "The necessary libraries for graphics (nilearn, matplotlib) "
            "were not found. Skipping graphics."
        )

    # If required, create surrogates, test, and export masked metrics
    if surr_type is not None:
        outprefix += "mkd_"
        LGR.info(
            f"Test significant {metric_name} against {n_surr} structurally "
            f"{surr_type} surrogates."
        )
        scgraph.create_surrogates(sc_type=surr_type, n_surr=n_surr, seed=seed)
        # #!# Export surrogates!
        if "sdi" in comp_metric or "gsdi" in comp_metric:
            scgraph.test_significance(method=method, p=p, return_masked=True)
            # Export thresholded metrics
            blocks.export_metric(scgraph, outext, outprefix)

            try:
                import matplotlib as _
                import nilearn as _

                if atlasname is not None:
                    LGR.info(f"Plot {metric_name} markerplot.")
                    if img is not None:
                        blocks.plot_metric(scgraph, outprefix, atlas=img, thr=0)
                    elif atlas is not None:
                        blocks.plot_metric(scgraph, outprefix, atlas=atlas, thr=0)

            except ImportError:
                pass

    LGR.info(f"End of workflow, find results in {outdir}.")

    return 0


def _main(argv=None):
    options = _get_parser().parse_args(argv)

    save_bash_call(options.fname, options.outdir, options.outname)

    nigsp(**vars(options))


if __name__ == "__main__":
    _main(sys.argv[1:])


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Operations for laplacian decomposition.

Attributes
----------
LGR
    Logger
"""

import logging
from copy import deepcopy

import numpy as np

LGR = logging.getLogger(__name__)


def compute_laplacian(mtx, negval="absolute", selfloops=False):
    """
    Compute Laplacian (L) matrix from a square matrix.

    mtx is supposed to be a connectivity matrix - its diagonal will be removed.
    L is obtained by subtracting the adjacency matrix from the degree matrix.

    Parameters
    ----------
    mtx : numpy.ndarray
        A square matrix
    negval : "absolute", "remove", or "rescale"
        The intended behaviour to deal with negative values in matrix:
        - "absolute" will take absolute values of the matrix
        - "remove" will set all negative elements to 0
        - "rescale" will rescale the matrix between 0 and 1.
        Default is "absolute".
    selfloops : "degree", bool, or numpy.ndarray
        Allow or remove self-loops in input matrix. A numpy array can be used to specify
        particular loops directly in the adjacency matrix.
        The degree matrix of the Adjacency matrix can also be used instead.
        In the last two cases, the degree matrix will be updated accordingly.
        Default is to remove self loops (False).

    Returns
    -------
    numpy.ndarray
        The laplacian of mtx
    numpy.ndarray
        The degree matrix of mtx as a (mtx.ndim-1)D array, updated with selfloops in case.

    See Also
    --------
    https://en.wikipedia.org/wiki/Laplacian_matrix

    Raises
    ------
    NotImplementedError
        If negval is not "absolute", "remove", or "rescale"
        If selfloop
    """
    mtx = deepcopy(mtx)
    if mtx.min() < 0:
        if negval == "absolute":
            mtx = abs(mtx)
        elif negval == "remove":
            mtx[mtx < 0] = 0
        elif negval == "rescale":
            mtx = (mtx - mtx.min()) / mtx.max()
        else:
            raise NotImplementedError(
                f'Behaviour "{negval}" to deal with negative values is not supported'
            )

    adjacency = deepcopy(mtx)

    if selfloops is False:
        adjacency[np.diag_indices(adjacency.shape[0])] = 0
    elif selfloops is True:
        pass
    elif type(selfloops) is np.ndarray:
        if selfloops.ndim > 1:
            raise NotImplementedError(
                "Multidimensional arrays are not implemented to specify self-loops"
            )
        if selfloops.shape[0] != mtx.shape[0]:
            raise ValueError(
                f"Array specified for self-loops has {selfloops.shape[0]} elements, "
                f"but specified matrix has {mtx.shape[0]} diagonal elements."
            )
        adjacency[np.diag_indices(adjacency.shape[0])] = selfloops
    elif selfloops == "degree":
        adjacency[np.diag_indices(adjacency.shape[0])] = 0
        adjacency[np.diag_indices(adjacency.shape[0])] = adjacency.sum(axis=1)
    else:
        raise NotImplementedError(
            f'Value "{selfloops}" for self-loops settings is not supported'
        )

    degree = adjacency.sum(axis=1)  # This is fixed to across columns

    degree_mat = np.zeros_like(mtx)
    degree_mat[np.diag_indices(degree_mat.shape[0])] = degree

    return degree_mat - adjacency, degree


def normalisation(lapl, degree, norm="symmetric", fix_zeros=True):
    """
    Normalise a Laplacian (L) matrix using either symmetric or random walk normalisation.

    Parameters
    ----------
    lapl : numpy.ndarray
        A square matrix that is a Laplacian, or a stack of Laplacian matrices.
    degree : np.ndarray or None, optional
        An array, a diagonal matrix, or a stack of either. This will be used as the
        the degree matrix for the normalisation.
        It's assumed that degree.ndim == lapl.ndim or degree.ndim == lapl.ndim-1.
    norm : ["symmetric", "symm", "random walk", "rw", random walk inflow", "rwi", "random walk outflow", "rwo"], str, optional
        The type of normalisation to perform. Default to symmetric.
        - "symmetric": D^(-1/2) @ L @ ^(-1/2), a.k.a. symmetric laplacian noramlisation
        - "random walk", "random walk inflow": D^(-1) @ L, a.k.a. random walk
          It normalises the inflow, i.e. it is row-optimised (each row = 0).
          Normally used in e.g. consensus networks.
        - "random walk outflow": L @ D^(-1)
          It normalises the outflow, i.e. it is column-optimised (each column = 0).
          Normally used in e.g. physical distribution networks.
    fix_zeros : bool, optional
        Whether to change 0 elements in the degree matrix to 1 to avoid multiplying by 0.
        Default is to do so.

    Returns
    -------
    numpy.ndarray
        The normalised laplacian

    Raises
    ------
    NotImplementedError
        If `lapl.ndim` - `degree.ndim` > 1
        If "norm" is not supported.
    ValueError
        If `d` in not a diagonal matrix or an array
        If `d` and `mtx` have different shapes.

    See Also
    --------
    https://en.wikipedia.org/wiki/Laplacian_matrix
    """
    deg = deepcopy(degree)
    if lapl.ndim - deg.ndim > 1:
        raise NotImplementedError(
            f"The provided degree matrix is {deg.ndim}D while the "
            f"provided laplacian matrix is {lapl.ndim}D."
        )
    elif lapl.ndim == deg.ndim:
        if not (deg.diagonal() == deg.sum(axis=1)).all():
            raise ValueError(
                "The provided degree matrix is not a diagonal matrix (or a stack of)."
            )
        deg = deepcopy(deg.diagonal())

    if deg.shape != lapl.shape[:-1]:
        raise ValueError(
            f"The provided degree matrix has shape {deg.shape} while the "
            f"provided matrix has shape {lapl.shape}."
        )

    if fix_zeros:
        deg[deg == 0] = 1

    d = np.zeros_like(lapl)

    if norm in ["symmetric", "symm"]:
        d[np.diag_indices(d.shape[0])] = deg ** (-1 / 2)
        return d @ lapl @ d
    elif norm in ["random walk", "rw", "random walk inflow", "rwi"]:
        d[np.diag_indices(d.shape[0])] = deg ** (-1)
        return d @ lapl
    elif norm in ["random walk outflow", "rwo"]:
        d[np.diag_indices(d.shape[0])] = deg ** (-1)
        return lapl @ d
    else:
        raise NotImplementedError(f'Normalisation type "{norm}" is not supported.')


def symmetric_normalised_laplacian(mtx, d=None, fix_zeros=True):
    """
    Compute symmetric normalised Laplacian (SNL) matrix.

    The SNL is obtained by pre- and post- multiplying mtx by the square root of the
    inverse of the diagonal and subtract the result from the identity matrix.
    Alternatively, it is possible to specify a different diagonal to do so.
    With zero-order nodes, the diagonals will contain 0s, returning a Laplacian
    with NaN elements. To avoid that, 0 elements in d will be changed to 1.

    Parameters
    ----------
    mtx : numpy.ndarray
        A [structural] matrix
    d : np.ndarray or None, optional
        Either an array or a diagonal matrix. If specified, d will be used as the
        **normalisation factor** (not the degree matrix).
    fix_zeros : bool, optional
        Whether to change 0 elements in the degree matrix to 1 to avoid multiplying by 0

    Returns
    -------
    numpy.ndarray
        The symmetric normalised laplacian of mtx

    Raises
    ------
        ValueError
            If `d` in not a diagonal matrix or an array
            If `d` and `mtx` have different shapes.

    Note
    ----
    This is here mainly for tests and legacy code, but it would be better not to use it!

    See Also
    --------
    https://en.wikipedia.org/wiki/Laplacian_matrix#Symmetrically_normalized_Laplacian_2
    """
    if d is not None:
        if d.ndim == 1:
            d = np.diag(d)
        elif not (np.diag(d) == d.sum(axis=-1)).all():
            raise ValueError(
                "The provided matrix for symmetric normalisation "
                "is not a diagonal matrix."
            )

        if d.shape != mtx.shape:
            raise ValueError(
                f"The provided diagonal has shape {d.shape} while the "
                f"provided matrix has shape {mtx.shape}."
            )

    colsum = mtx.sum(axis=-1)

    identity_mat = np.zeros_like(colsum)
    identity_mat[colsum != 0] = 1
    identity_mat = np.diag(identity_mat)

    if fix_zeros:
        colsum[colsum == 0] = 1
    if d is None:
        d = np.diag(colsum ** (-1 / 2))

    symm_norm = d @ mtx @ d

    return identity_mat - symm_norm
    # ## Fix Identity matrix by giving


def decomposition(mtx):
    """
    Run a eigenvector decomposition on input.

    Parameters
    ----------
    mtx : numpy.ndarray
        The matrix to decompose

    Returns
    -------
    numpy.ndarray
        The eigenvalues resulting from the decomposition
    numpy.ndarray
        The eigenvectors resulting from the decomposition
    """
    eigenval, eigenvec = np.linalg.eig(mtx)

    idx = np.argsort(eigenval)
    eigenval = eigenval[idx]
    # #!# Check that eigenvec has the right index and not inverted
    eigenvec = eigenvec[:, idx]

    return eigenval, eigenvec


def recomposition(eigenval, eigenvec):
    """
    Recompose a matrix from its eigenvalues and eigenvectors.

    At the moment, it supports only 2D (not stacks).

    Parameters
    ----------
    eigenval : numpy.ndarray
        Array of eigenvalues. The program detects if it's a diagonal matrix or not.
    eigenvec : numpy.ndarray
        Matrix of eigenvectors.

    Returns
    -------
    numpy.ndarray
        The reconstructed matrix
    """
    if eigenvec.ndim > 2:
        raise NotImplementedError(
            f"Given matrix dimensionality ({eigenvec.ndim}) is not supported."
        )

    if eigenval.ndim == eigenvec.ndim - 1:
        eigenval = np.diag(eigenval)
    elif eigenval.ndim < eigenvec.ndim - 1:
        raise ValueError("Not enough dimensions in given eigenvalue matrix.")
    elif eigenval.ndim > eigenvec.ndim:
        raise ValueError("Too many dimensions in given eigenvalue matrix.")
    elif not (np.diag(eigenval) == eigenval.sum(axis=-1)).all():
        raise ValueError("The provided eigenvalue matrix is not a diagonal matrix.")

    mtx = eigenvec @ eigenval @ eigenvec.T

    return mtx


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for operations.timeseries."""

import numpy as np
from numpy.random import rand
from pytest import mark, raises

from nigsp.operations import timeseries
from nigsp.utils import prepare_ndim_iteration


# ### Unit tests
def test_normalise_ts():
    ts = np.arange(24).reshape(2, 3, 4, order="F")

    z = np.asarray([[-1, 0, 1], [-1, 0, 1]], dtype="float64")
    z = np.repeat(z[..., np.newaxis], 4, axis=-1)

    zg = np.arange(6).reshape(3, 2).T
    zg = (zg - zg.mean()) / zg.std(ddof=1)
    zg = np.repeat(zg[..., np.newaxis], 4, axis=-1)

    tsz = timeseries.normalise_ts(ts)

    tsgz = timeseries.normalise_ts(ts, globally=True)

    assert (tsz == z).all()
    assert (tsz.sum() - 0) < 1e-10
    assert (tsgz == zg).all()
    assert (tsgz.sum() - 0) < 1e-10

    ts = np.arange(4)
    assert (ts == timeseries.normalise_ts(ts)).all()


def test_spc_ts():
    ts = np.arange(24).reshape(2, 3, 4, order="F")
    spc = (ts - ts.mean(axis=1)[:, np.newaxis, ...]) / ts.mean(axis=1)[
        :, np.newaxis, ...
    ]
    spc[np.isnan(spc)] = 0

    spcg = (ts - ts.mean(axis=(0, 1))) / ts.mean(axis=(0, 1))
    spcg[np.isnan(spcg)] = 0

    tsspc = timeseries.spc_ts(ts)

    tsgspc = timeseries.spc_ts(ts, globally=True)

    assert (tsspc == spc).all()
    assert (tsspc.sum() - 0) < 1e-10
    assert (tsgspc == spcg).all()
    assert (tsgspc.sum() - 0) < 1e-10

    ts = np.arange(4)
    assert (ts == timeseries.spc_ts(ts)).all()


def test_demean_ts():
    ts = np.arange(24).reshape(2, 3, 4, order="F")

    dm = np.asarray([[-2, 0, 2], [-2, 0, 2]], dtype="float64")
    dm = np.repeat(dm[..., np.newaxis], 4, axis=-1)

    dmg = np.repeat(
        np.arange(-2.5, 2.6, 1).reshape(3, 2).T[..., np.newaxis], 4, axis=-1
    )

    tsdm = timeseries.demean_ts(ts)

    tsgdm = timeseries.demean_ts(ts, globally=True)

    assert (tsdm == dm).all()
    assert (tsgdm == dmg).all()

    ts = np.arange(4)
    assert (ts == timeseries.demean_ts(ts)).all()


def test_rescale_ts():
    ts = np.arange(24).reshape(2, 3, 4, order="F")

    res = np.asarray([[1, 1.5, 2], [1, 1.5, 2]], dtype="float64")
    res = np.repeat(res[..., np.newaxis], 4, axis=-1)

    resg = np.repeat(
        np.arange(1, 2.1, 0.2).round(1).reshape(3, 2).T[..., np.newaxis], 4, axis=-1
    )

    tsres = timeseries.rescale_ts(ts, vmin=1, vmax=2)

    tsgres = timeseries.rescale_ts(ts, vmin=1, vmax=2, globally=True)

    assert (tsres == res).all()
    assert (tsgres == resg).all()

    ts = np.arange(4)
    assert (ts == timeseries.rescale_ts(ts)).all()


def test_resize_ts():
    ts = np.arange(24).reshape(2, 3, 4, order="F")
    gsr = np.repeat(
        np.repeat(
            np.asarray([-0.5, 0.5], dtype="float64")[..., np.newaxis], 3, axis=-1
        )[..., np.newaxis],
        4,
        axis=-1,
    )

    tsgsr = timeseries.resize_ts(ts, resize="gsr")

    assert (tsgsr == gsr).all()
    assert (ts == timeseries.resize_ts(ts)).all()

    ts = np.arange(4)
    assert (ts == timeseries.resize_ts(ts)).all()


def test_graph_fourier_transform():
    ts = rand(3, 6)
    ev = rand(3, 3)
    pr = ev.conj().T @ ts

    proj = timeseries.graph_fourier_transform(ts, ev)

    assert (proj == pr).all()

    ts = rand(3, 6, 2, 2)
    tsr, pr = prepare_ndim_iteration(ts, 2)

    for i in range(tsr.shape[-1]):
        pr[:, :, i] = ev.conj().T @ np.squeeze(tsr[:, :, i])
    pr = pr.reshape(ts.shape)
    prm = pr.mean(axis=1)
    engm = (pr**2).mean(axis=1)

    proj = timeseries.graph_fourier_transform(ts, ev, mean=True)
    eng = timeseries.graph_fourier_transform(ts, ev, energy=True, mean=True)

    assert (proj == prm).all()
    assert (eng == engm).all()


def test_median_cutoff_frequency_idx():
    energy = np.array([[0, 3], [1, 2], [0, 1], [0.5, 0.5]])
    freq_t = 2

    freq_idx = timeseries.median_cutoff_frequency_idx(energy)

    assert freq_t == freq_idx


def test_graph_filter():
    ts = rand(4, 10)
    ev = rand(4, 4)

    keys = ["hi", "ho"]
    freq_idx = 2

    ev_s = dict.fromkeys(keys)
    ts_s = dict.fromkeys(keys)

    ev_s["hi"] = np.append(
        ev[:, :freq_idx], np.zeros_like(ev[:, freq_idx:], dtype="float32"), axis=-1
    )
    ev_s["ho"] = np.append(
        np.zeros_like(ev[:, :freq_idx], dtype="float32"), ev[:, freq_idx:], axis=-1
    )

    fcf = timeseries.graph_fourier_transform(ts, ev)
    ts_s["ho"] = timeseries.graph_fourier_transform(fcf, ev_s["ho"].T)
    ts_s["hi"] = timeseries.graph_fourier_transform(fcf, ev_s["hi"].T)

    evec_split, ts_split = timeseries.graph_filter(ts, ev, freq_idx, keys)

    assert all(item in keys for item in list(evec_split.keys()))
    assert all(item in keys for item in list(ts_split.keys()))
    for k in keys:
        assert (ev_s[k] == evec_split[k]).all()
        assert (ts_s[k] == ts_split[k]).all()


# ### Break tests
def test_break_resize_ts():
    with raises(NotImplementedError) as errorinfo:
        timeseries.resize_ts(np.arange(6).reshape(2, 3), resize=[2, 3, 4])
    assert "two elements" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        timeseries.resize_ts(np.arange(6).reshape(2, 3), resize="Jonah")
    assert "method Jonah" in str(errorinfo.value)


def test_break_median_cutoff_frequency_idx():
    with raises(NotImplementedError) as errorinfo:
        timeseries.median_cutoff_frequency_idx(rand(2, 3, 4))
    assert "have 3 dimensions" in str(errorinfo.value)


@mark.parametrize("freq", [(0), (2), (4)])
def test_break_graph_filter(freq):
    with raises(IndexError) as errorinfo:
        timeseries.graph_filter(rand(2, 3), rand(2, 2), freq)
    assert f"index {freq} is not valid" in str(errorinfo.value)



#!/usr/bin/env python3
"""Tests for operations.graph."""

import numpy as np

from nigsp.operations import graph


# ### Unit tests
def test_zerocross():
    """Test zerocross with legendre polynomials."""

    def _bonnet(d, x):
        if d == 0:
            return np.ones_like(x)
        elif d == 1:
            return x
        else:
            return (
                (2 * d - 1) * x * _bonnet(d - 1, x) - (d - 1) * _bonnet(d - 2, x)
            ) / d

    x = np.linspace(-1, 1, 100)
    legendre = np.empty([100, 5], dtype="float32")
    for n in range(5):
        legendre[:, n] = _bonnet(n, x)

    zx = np.linspace(0, 4, 5)

    assert all(zx == graph.zerocross(legendre))


def test_nodestrength():
    """Test nodestrength with random matrix."""
    a = np.random.rand(3, 3)
    a = a - a.mean()
    b = np.abs(a)
    s = b.sum(axis=0)
    f = s.mean(axis=-1)

    n = graph.nodestrength(a)
    m = graph.nodestrength(b)
    o = graph.nodestrength(a, mean=True)

    assert all(n == m)
    assert all(n == s)
    assert f == o



#!/usr/bin/env python3
"""
Surrogate utility.

Attributes
----------
LGR
    Logger
"""

import logging
from copy import deepcopy
from math import ceil, factorial, floor

import numpy as np

from nigsp.operations.laplacian import decomposition
from nigsp.operations.timeseries import graph_fourier_transform

LGR = logging.getLogger(__name__)
SURR_TYPE = ["informed", "uninformed"]
STAT_METHOD = ["Bernoulli", "frequentist"]


def random_sign(eigenvec, n_surr=1000, seed=42, stack=False):
    """
    Create surrogates by randomly switching signs of eigenvectors.

    Parameters
    ----------
    eigenvec : numpy.ndarray
        A matrix of eigenvectors
    n_surr : int, optional
        Number of surrogates to create
    seed : int or None, optional
        Random seed (for repeatability)
    stack : bool, optional
        If True, add original eigenvec as last entry of the last dimension
        of the created surrogate matrix

    Returns
    -------
    numpy.ndarray
        The matrix of surrogates, of shape eigenvec * n_surr(+1)

    Raises
    ------
    NotImplementedError
        If eigenvec is 4+ D.
    """
    # Reinitialise the random seed for repeatability
    rng = np.random.default_rng(seed)

    if eigenvec.ndim > 3:
        raise NotImplementedError(
            f"Provided data has {eigenvec.ndim} dimensions, "
            "but data of more than 3 dimensions are not "
            "supported yet"
        )

    rand_evec = np.empty((eigenvec.shape + (n_surr,)), dtype="float32")

    LGR.info("Randomly switching signs of eigenvectors to create surrogates.")
    for i in range(n_surr):
        # #!# Check if two conditions can be merged.
        if eigenvec.ndim < 3:
            r_sign = rng.integers(0, 1, eigenvec.shape[0], endpoint=True)
            r_sign[r_sign == 0] = -1
            rand_evec[..., i] = eigenvec * r_sign
        else:
            for j in range(eigenvec.shape[2]):
                r_sign = rng.integers(0, 1, eigenvec.shape[0], endpoint=True)
                r_sign[r_sign == 0] = -1
                rand_evec[:, :, j, i] = eigenvec[..., j] * r_sign

    if stack:
        rand_evec = np.append(rand_evec, eigenvec[..., np.newaxis], axis=-1)

    return rand_evec


def _create_surr(timeseries, eigenvec, n_surr, seed, stack):
    """
    Proper surrogate creation step.

    This is not meant to be called as a function.

    Parameters
    ----------
    timeseries : numpy.ndarray
        A 3D (or less) array coding a timeseries in the second axis (axis 1).
    eigenvec : numpy.ndarray
        The eigenvector matrix from a previous Laplacian decomposition.
    n_surr : int
        The number of surrogates to create
    seed : int or None
        The seed to reinitialise the RNG - used for replicability.
    stack : bool
        If True, append the real matrix at the end of the stack.

    Returns
    -------
    numpy.ndarray
        The surrogate matrix, of shape timeseries.shape, n_surr

    Raises
    ------
    NotImplementedError
        If timeseries is 4+ D and/or eigenvector matrix has not enough dimensions.
    """
    rand_evec = random_sign(eigenvec, n_surr, seed, stack)

    if stack:
        n_surr += 1

    surr = np.empty((timeseries.shape + (n_surr,)), dtype="float32")

    fourier_coeff = graph_fourier_transform(timeseries, eigenvec)

    LGR.info("Projecting the timeseries onto the surrogate eigenvectors.")
    for i in range(n_surr):
        if timeseries.ndim < 3 and rand_evec.ndim == timeseries.ndim + 1:
            surr[..., i] = graph_fourier_transform(fourier_coeff, rand_evec[..., i].T)
        elif timeseries.ndim == 3:
            if rand_evec.ndim < 4:
                surr[..., i] = graph_fourier_transform(
                    fourier_coeff, rand_evec[..., i].T
                )
            else:
                for j in range(rand_evec.shape[2]):
                    surr[:, :, j, i] = graph_fourier_transform(
                        fourier_coeff, rand_evec[:, :, j, i].T
                    )
        else:
            raise NotImplementedError(
                "No solution implemented for timeseries "
                f"of {timeseries.ndim} dimension(s) and "
                f"eigenvector matrix of {eigenvec.ndim} "
                "dimension(s)"
            )
    return surr


def sc_informed(timeseries, eigenvec, n_surr=1000, seed=124, stack=False):
    """
    Create surrogates informed by the real structural connectivity.

    Parameters
    ----------
    timeseries : numpy.ndarray
        A 3D (or less) array coding a timeseries in the second axis (axis 1).
    eigenvec : numpy.ndarray
        The eigenvector matrix from a previous Laplacian decomposition.
    n_surr : int, optional
        The number of surrogates to create
    seed : int or None, optional
        The seed to reinitialise the RNG - used for replicability.
    stack : bool, optional
        If True, append the real matrix at the end of the stack.

    Returns
    -------
    numpy.ndarray
        The surrogate matrix, of shape timeseries.shape, n_surr

    Raises
    ------
    NotImplementedError
        If timeseries is 4+ D.
    """
    if timeseries.ndim > 3:
        raise NotImplementedError(
            f"Provided timeseries has {timeseries.ndim} "
            "dimensions, but timeseries of more than 3 "
            "dimensions are not supported yet."
        )

    return _create_surr(timeseries, eigenvec, n_surr, seed, stack)


def sc_uninformed(timeseries, lapl_mtx, n_surr=1000, seed=98, stack=False):
    """
    Create surrogates ignorant of the real structural connectivity.

    Parameters
    ----------
    timeseries : numpy.ndarray
        A 3D (or less) array coding a timeseries in the second axis (axis 1).
    lapl_mtx : numpy.ndarray
        A symmetrically normalised laplacian matrix.
    n_surr : int, optional
        The number of surrogates to create
    seed : int or None, optional
        The seed to reinitialise the RNG - used for replicability.
    stack : bool, optional
        If True, append the real matrix at the end of the stack.

    Returns
    -------
    numpy.ndarray
        The surrogate matrix, of shape timeseries.shape, n_surr

    Raises
    ------
    NotImplementedError
        If timeseries is 4+ D.
    """
    if timeseries.ndim > 3:
        raise NotImplementedError(
            f"Provided timeseries has {timeseries.ndim} "
            "dimensions, but timeseries of more than 3 "
            "dimensions are not supported yet."
        )

    symm_norm = np.eye(lapl_mtx.shape[0]) - lapl_mtx
    symm_norm_sum = symm_norm.sum(axis=-1)

    conf_model = np.outer(symm_norm_sum, symm_norm_sum.T) / symm_norm.sum()

    conf_lapl = np.diag(symm_norm_sum) - conf_model

    _, surr_eigenvec = decomposition(conf_lapl)

    return _create_surr(timeseries, surr_eigenvec, n_surr, seed, stack)


def test_significance(
    surr,
    data=None,
    method="Bernoulli",
    p=0.05,
    p_bernoulli=None,
    return_masked=False,
    mean=False,
):
    """
    Test the significance of the empirical data against surrogates.

    Two methods are implemented, 'Bernoulli' and 'frequentist'.
    - 'frequentist' is a group or single subject test. It tests that the
      empirical data are in the highest (or lowest) percentile (where the
      percentile is defined by p/2).
    - 'Bernoulli' is a group test. It tests that the number of subjects for
      which the empirical data is higher (or lower) than a set of surrogates
      (frequentist approach) is at the tail of a binomial cumulative
      distribution (where 'tail' is defined by p).

    Note that p is expressed as two-tails test for the frequentist approach and
    a one-tail test for the Bernoulli approach.

    Both surr and data are expected to have first dimensions: observations [x subjects].

    Parameters
    ----------
    surr : numpy.ndarray
        The surrogate matrix, where all surrogates are aligned along the last axis.
        May have the empirical data matrix last along the last axis.
        Expected to have shape: observations, [subjects,] surrogates.
    data : numpy.ndarray or None, optional
        The empirical data matrix. If given, it's appended at the end of the
        surrogate matrix.
        Expected to have shape: observations[, subjects].
    method : 'Bernoulli' or 'frequentist', optional
        The method to adopt for testing, either based on a Bernoulli process
        or a frequentist observation (see above).
    p : float, optional
        The probability threshold to adopt for the frequentist approach part.
        Note that this is a two-tails test.
    p_bernoulli : float or None, optional
        The probability threshold to adopt for Bernoulli's test.
        If left as None, the specified p value will be used instead,
        and p will be set to 0.1.
        Note that this is a one-tail test.
    return_masked : bool, optional
        If True, returns the masked data. If False, returns a mask that holds
        True where the good data are (inverse of numpy mask). Mask has the same
        shape as data.
    mean : bool, optional
        If True, returns the average of the masked data along the last axis.

    Returns
    -------
    numpy.ndarray
        A numpy.ndarray shaped observations[, subjects]. If return_masked is True,
        returns the masked version of `data`, otherwise returns the mask.
        If mean is True, returns the average along the subject axis.

    Raises
    ------
    ValueError
        If data is not None and the surrogate shape (except last axis) is
        different from the data shape
    NotImplementedError
        If any other method rather than those listed above is selected.

    """
    # #!# Check that the surrogate shape has parcels in the first axis!
    # If provided, append data to surr
    if data is not None:
        if surr.shape[: data.ndim] != data.shape:
            raise ValueError(
                "Provided empirical data and surrogate data shapes "
                f"do not agree, with shapes {data.shape} and "
                f"{surr.shape[:data.ndim]} (last axis excluded)"
            )
        if not (surr[..., -1] == data).all():
            # Check that data was not appended yet.
            surr = np.append(surr, data[..., np.newaxis], axis=-1)

    if p < 0 or p > 1:
        raise ValueError(
            "p values should always be between 0 and 1. The "
            f"provided value of {p} is out of these boundaries"
        )
    elif p == 0 or p == 1:
        LGR.warning(
            f"The selected p value of {p} is at the limits of the "
            "possible range of [0, 1]. Statistical thresholding might "
            "not be interpretable."
        )

    if surr.ndim < 3:
        LGR.warning(
            f"Warning: surrogate dimensions ({surr.ndim}) are less than "
            "the program expects - check that you mean to run a test on "
            "an average or that you have enough surrogates."
        )

    # Reorder the surrogate matrix, then find where the real surrogate is
    LGR.info("Reordering surrogates for test")
    real_idx = surr.shape[-1] - 1
    reord_surr = np.argsort(surr, axis=-1) == real_idx

    LGR.info(f"Adopting {method} testing method.")
    # Testing both tails requires to split p
    if method == "frequentist":
        LGR.info(f"Testing for p={p} two-tails (p={p/2} each tail)")
        p = p / 2
        # If there aren't enough surrogates, send a warning message on the real p
        # Then update p
        if 1 / surr.shape[-1] > p:
            LGR.warning(
                "The generated surrogates are not enough to test for "
                f"the selected p ({p*2} two-tails), since at least "
                f"{ceil(1/p)-1} surrogates are required for the selected "
                f"p value. Testing for p={1/surr.shape[-1]} two-tails instead."
            )
            p = 1 / surr.shape[-1]

    elif method == "Bernoulli":
        # If there aren't enough subjects, send a warning message on the real p
        # Then update group level p
        if p_bernoulli is None:
            p_bernoulli = deepcopy(p)
            p = 0.1
        if 1 / surr.shape[1] > p_bernoulli:
            LGR.warning(
                "The provided subjects are not enough to test for "
                f"p={p_bernoulli} one-tail at the group level, since "
                f"at least {ceil(1/p_bernoulli)} subjects are required."
            )
            p_bernoulli = 1 / surr.shape[1]
        # If there aren't enough surrogates, send a warning message on the real p
        # Then update subject level p
        if 1 / surr.shape[-1] > p:
            LGR.warning(
                "The generated surrogates are not enough to test for "
                f"p={p} two-tails at the subject level. "
                f"{ceil(1/p)-1} surrogates are required for p={p}."
            )
            p = 1 / surr.shape[-1]

        LGR.info(
            f"Testing for p={p_bernoulli} one-tail at the group level and "
            f"at p={p*2} two-tails (p={p} each tail) at the subject level."
        )
    else:
        raise NotImplementedError(
            "Other testing methods than Bernoulli or "
            "frequentist are not implemented at the moment."
        )

    # First, and no matter what, apply frequentist approach to find where
    # the real data index (real_idx) is at the extremes of the matrix last axis
    # (with tolerance on the extremes depending on p).
    # real_idx serendipitously is the number of surrogates.
    stat_mask = reord_surr[..., : floor(real_idx * p) + 1].any(axis=-1) + reord_surr[
        ..., -floor(real_idx * p) - 1 :
    ].any(axis=-1)

    if method == "Bernoulli" and surr.shape[1] > 1 and surr.ndim >= 3:
        # The following computes the CDF of a binomial distribution
        # Difference with scipy's binom.cdf (100 samples) is: 5.066394802133445e-06
        # #!# See if there is a quicker way to get this (probably invert testing)

        def _pmf(x, n, p):
            f = (
                (factorial(n) / (factorial(x) * factorial(n - x)))
                * p**x
                * (1 - p) ** (n - x)
            )
            return f

        x = np.arange(0, 100, 1)
        # Generate the PMF of the binomial distribution
        y = np.asarray([_pmf(i, 100, p) for i in x], dtype="float32")
        # Generate the CDF, then invert it.
        y = 1 - np.cumsum(y)
        # Find the number of subjects necessary to be statistically significant,
        # adjusted for the number of subjects in the surrogates.
        # Then find all parcels for which the real data is higher or lower
        # than all surrogates in enough subjects.
        # The +1 in thr is to be conservative on the number of subjects.
        # p_bernoulli/surr.shape[0] is a Bonferroni correction.
        thr = x[y < p_bernoulli / surr.shape[0]][0]
        thr = np.floor(surr.shape[1] / 100 * thr) + 1
        # On top of the frequentist approach, find the parcels that pop up
        # in the frequentist approach for enough subjects.
        stat_mask = stat_mask.sum(axis=1) > thr
        # repeat stat_mask for the number of subjects.
        stat_mask = stat_mask[..., np.newaxis].repeat(surr.shape[1], axis=-1)
    elif surr.shape[1] == 1 and surr.ndim >= 3:
        LGR.warning(
            'The "Bernoulli" method is a group test that requires '
            "multiple subjects to be run."
        )
    elif surr.ndim < 3:
        LGR.warning(
            "The dimensionality of the data is not enough to run "
            'the "Bernoulli" method.'
        )

    if return_masked:
        LGR.info("Returning masked empirical data")
        stat_mask = np.ma.array(
            data=surr[..., -1], mask=np.invert(stat_mask), fill_value=np.NINF
        ).squeeze()
    else:
        LGR.info("Returning mask")

    if mean and stat_mask.ndim >= 2:
        LGR.info("Returning average across subjects (axis 1)")
        stat_mask = stat_mask.mean(axis=1)

    if return_masked:
        stat_mask = stat_mask.filled()

    return stat_mask


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for operations.surrogates."""
import numpy as np
from numpy.random import default_rng, rand
from pytest import raises

from nigsp.operations import decomposition, graph_fourier_transform, surrogates


# ### Unit tests
def test_random_sign():
    n_surr = 4
    random_seed = 2
    eigenvec = rand(20)

    rs = surrogates.random_sign(eigenvec, n_surr, random_seed, stack=True)

    # For the moment, replicate code inside function.
    rng = default_rng(random_seed)
    rand_evec = np.empty((eigenvec.shape + (n_surr,)), dtype="float32")
    for i in range(n_surr):
        r_sign = rng.integers(0, 1, eigenvec.shape[0], endpoint=True)
        r_sign[r_sign == 0] = -1
        rand_evec[..., i] = eigenvec * r_sign
    rand_evec = np.append(rand_evec, eigenvec[..., np.newaxis], axis=-1)

    assert rs.ndim == 2
    assert rs.shape == eigenvec.shape + (n_surr + 1,)
    assert (rand_evec == rs).all()

    # #!# Implement stricter checks on signs
    pass


def test_create_surr():
    eigenvec = rand(2, 2)
    timeseries = rand(2, 4)
    n_surr = 3
    seed = 2

    cs = surrogates._create_surr(timeseries, eigenvec, n_surr, seed, stack=True)

    rand_evec = surrogates.random_sign(eigenvec, n_surr, seed, stack=True)
    surr = np.empty((timeseries.shape + (n_surr + 1,)), dtype="float32")
    fourier_coeff = graph_fourier_transform(timeseries, eigenvec)

    for i in range(n_surr + 1):
        surr[..., i] = graph_fourier_transform(fourier_coeff, rand_evec[..., i].T)

    assert cs.ndim == 3
    assert cs.shape == surr.shape
    assert (cs == surr).all()


def test_sc_informed():
    eigenvec = rand(2, 2)
    timeseries = rand(2, 4)
    n_surr = 3
    seed = 2

    surr = surrogates._create_surr(timeseries, eigenvec, n_surr, seed, stack=True)
    cs = surrogates.sc_informed(timeseries, eigenvec, n_surr, seed, stack=True)

    assert cs.ndim == 3
    assert cs.shape == surr.shape
    assert (cs == surr).all()


def test_sc_uninformed():
    lapl_mtx = rand(2, 2)
    timeseries = rand(2, 4)
    n_surr = 3
    seed = 2

    symm_norm = np.eye(lapl_mtx.shape[0]) - lapl_mtx
    symm_norm_sum = symm_norm.sum(axis=-1)
    conf_model = np.outer(symm_norm_sum, symm_norm_sum.T) / symm_norm.sum()
    conf_lapl = np.diag(symm_norm_sum) - conf_model
    _, surr_eigenvec = decomposition(conf_lapl)
    surr = surrogates._create_surr(timeseries, surr_eigenvec, n_surr, seed, stack=True)

    cs = surrogates.sc_uninformed(timeseries, lapl_mtx, n_surr, seed, stack=True)

    assert cs.ndim == 3
    assert cs.shape == surr.shape
    assert (cs == surr).all()


def test_test_significance():
    # Create surrogates so that one subject has 3 significant observations over 4,
    # and in group 1 observation is significant over 5
    surr = np.array(
        [
            [
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.1, 0.1, 0.1, 0.1, -0.1],
                [0.4, 0.4, 0.1, 0.1, 0.2],
            ],
            [
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
            ],
            [
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
            ],
            [
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
            ],
            [
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.1, 0.1, 0.1, 0.1, 0.4],
                [0.4, 0.4, 0.1, 0.1, 0.2],
                [0.4, 0.4, 0.1, 0.1, 0.2],
            ],
        ]
    )

    data = surr[..., -1]
    p_bernoulli = 0.35
    p = 0.4
    # Test frequentist method first
    mask_freq = np.array(
        [
            [True, True, True, False],
            [True, True, False, False],
            [True, True, False, False],
            [False, False, False, False],
            [True, True, False, False],
        ]
    )
    mask = surrogates.test_significance(surr, data, method="frequentist", p=p)

    assert (mask_freq == mask).all()

    # Test Bernoulli's method after (mask created flat in 1D for convenience)
    mask_bern = np.array([True, False, False, False, False])
    mask_bern = mask_bern[..., np.newaxis].repeat(surr.shape[1], axis=-1)
    mask = surrogates.test_significance(surr, p=p, p_bernoulli=p_bernoulli)

    assert (mask_bern == mask).all()

    # Test masked and average
    masked_data = np.ma.array(data=data, mask=np.invert(mask_freq), fill_value=0)
    mask = surrogates.test_significance(
        surr, method="frequentist", p=p, return_masked=True
    )

    assert (masked_data == mask).all()

    masked_data_avg = masked_data.mean(axis=1)
    mask_avg = surrogates.test_significance(
        surr, method="frequentist", p=p, return_masked=True, mean=True
    )

    assert (masked_data_avg == mask_avg).all()


# ### Break tests
def test_break_random_sign():
    with raises(NotImplementedError) as errorinfo:
        surrogates.random_sign(rand(2, 3, 4, 5))
    assert "has 4 dimensions" in str(errorinfo.value)


def test_break_create_surr():
    with raises(NotImplementedError) as errorinfo:
        surrogates._create_surr(rand(2, 3, 4, 5), rand(2, 2), 1, 2, False)
    assert "timeseries of 4 dimension(s)" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        surrogates._create_surr(rand(2), rand(2, 2), 1, 2, False)
    assert (
        "timeseries of 1 dimension(s) and eigenvector matrix of 2 dimension(s)"
        in str(errorinfo.value)
    )


def test_break_sc_informed():
    with raises(NotImplementedError) as errorinfo:
        surrogates.sc_informed(rand(2, 3, 4, 5), rand(2, 2))
    assert "has 4 dimensions" in str(errorinfo.value)


def test_break_sc_uninformed():
    with raises(NotImplementedError) as errorinfo:
        surrogates.sc_uninformed(rand(2, 3, 4, 5), rand(2, 2))
    assert "has 4 dimensions" in str(errorinfo.value)


def test_break_test_significance():
    with raises(ValueError) as errorinfo:
        surrogates.test_significance(rand(2, 3, 5), rand(2, 4))
    assert "shapes do not agree" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        surrogates.test_significance(rand(2, 4, 3), rand(2, 4), p=-0.1)
    assert "p values should always be between 0 and 1" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        surrogates.test_significance(rand(2, 4, 3), rand(2, 4), p=1.1)
    assert "p values should always be between 0 and 1" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        surrogates.test_significance(rand(2, 3, 5), method="Baobab")
    assert "Other testing methods" in str(errorinfo.value)



"""Hopefully importing everything."""

import pkgutil

from ._version import get_versions
from .operations import graph, laplacian, metrics, nifti, surrogates, timeseries

SKIP_MODULES = ["tests"]

__version__ = get_versions()["version"]
del get_versions

__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    if "tests" not in module_name:
        __all__.append(module_name)
        _module = loader.find_module(module_name).load_module(module_name)
        globals()[module_name] = _module



#!/usr/bin/env python3
"""
Blocks of code for workflows.

Attributes
----------
LGR
    Logger
"""

import logging

from nigsp import io, viz
from nigsp.operations import nifti

LGR = logging.getLogger(__name__)


def nifti_to_timeseries(fname, atlasname):
    """
    Read a nifti file and returns a normalised timeseries from an atlas.

    Parameters
    ----------
    fname : string or os.PathLike
        Filename (and path) of a functional timeseries nifti dataset.
    atlasname : string or os.PathLike
        Filename (and path) of an atlas nifti dataset.

    Returns
    -------
    numpy.ndarray
        The average parcel timeseries from the given atlas.
    numpy.ndarray
        The atlas data.
    3D Nifti1Image
        The full nifti image of the atlas.
    """
    _, fname = io.check_ext(io.EXT_NIFTI, fname, scan=True)
    _, atlasname = io.check_ext(io.EXT_NIFTI, atlasname, scan=True)

    data, mask, _ = io.load_nifti_get_mask(fname)
    atlas, amask, img = io.load_nifti_get_mask(atlasname, is_mask=True, ndim=3)
    mask *= amask

    timeseries = nifti.apply_atlas(data, atlas, mask)

    return timeseries, atlas, img


def export_metric(scgraph, outext, outprefix):
    """
    Export the metrics computed within the library.

    Parameters
    ----------
    scgraph : SCGraph object
        The native object of this library.
    outext : str
        The desired extension for export - it will force the type of file created.
    outprefix : str
        The desired prefix for the export.

    Returns
    -------
    0
        If everything goes well
    """
    if outext in io.EXT_NIFTI:
        try:
            import nibabel as _
        except ImportError:
            LGR.warning(
                "The necessary library for nifti export (nibabel) "
                "was not found. Exporting metrics in TSV format instead."
            )
            outext = ".tsv.gz"
        if scgraph.img is None:
            LGR.warning(
                "A necessary atlas nifti image was not found. "
                "Exporting metrics in TSV format instead."
            )
            outext = ".tsv.gz"

    if scgraph.sdi is not None:
        if outext in io.EXT_NIFTI:
            data = nifti.unfold_atlas(scgraph.sdi, scgraph.atlas)
            io.export_nifti(data, scgraph.img, f"{outprefix}sdi{outext}")
        else:
            io.export_mtx(scgraph.sdi, f"{outprefix}sdi{outext}")
    elif scgraph.gsdi is not None:
        for k in scgraph.gsdi.keys():
            if outext in io.EXT_NIFTI:
                data = nifti.unfold_atlas(scgraph.gsdi[k], scgraph.atlas)
                io.export_nifti(data, scgraph.img, f"{outprefix}gsdi_{k}{outext}")
            else:
                io.export_mtx(scgraph.gsdi[k], f"{outprefix}gsdi_{k}{outext}")

    return 0


def plot_metric(scgraph, outprefix, atlas=None, thr=None):
    """
    If possible, plot metrics as markerplot.

    Parameters
    ----------
    scgraph : SCGraph object
        The internal object containing all data.
    outprefix : str
        The prefix of the png file to export
    img : 3DNiftiImage or None, optional
        The nifti image of the atlas
    atlas : 3D Nifti1Image, numpy.ndarray, or None, optional
        Either a nifti image containing a valid atlas or a set of parcel coordinates.
    thr : float or None, optional
        The threshold to use in plotting the nodes.
    """
    # Check that atlas format is supported.
    try:
        atlas.header
        atlas_plot = atlas
    except AttributeError:
        try:
            atlas.min()
            if atlas.ndim == 2 and atlas.shape[1] == 3:
                atlas_plot = atlas
        except AttributeError:
            LGR.warning(
                "The provided atlas is not in a format supported for markerplots."
            )
            atlas_plot = None

    # If it is, plot.
    if atlas_plot is not None:
        if scgraph.sdi is not None:
            viz.plot_nodes(
                scgraph.sdi, atlas_plot, filename=f"{outprefix}sdi.png", thr=thr
            )
        elif scgraph.gsdi is not None:
            for k in scgraph.gsdi.keys():
                viz.plot_nodes(
                    scgraph.gsdi[k],
                    atlas_plot,
                    filename=f"{outprefix}gsdi_{k}.png",
                    thr=thr,
                )

    return 0


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



"""References to be imported and injected throughout the package."""
from nigsp.due import Doi

PRETI_2019 = Doi("10.1038/s41467-019-12765-7")

GRIFFA_2022 = Doi("10.1016/j.neuroimage.2022.118970")

NIGSP = Doi("10.5281/zenodo.6373436")



#!/usr/bin/env python3
"""Tests for blocks."""
import shutil
import sys
from os import makedirs, remove
from os.path import isfile, join

import nibabel
import numpy as np
from nilearn.plotting import find_parcellation_cut_coords
from pymatreader import read_mat
from pytest import mark

from nigsp import blocks
from nigsp.io import load_nifti_get_mask
from nigsp.objects import SCGraph


# ### Unit tests
def test_nifti_to_timeseries(atlastime, atlas):
    img_in = nibabel.load(atlas)
    atlas_in = img_in.get_fdata()
    ts_out, atlas_out, img_out = blocks.nifti_to_timeseries(atlastime, atlas)
    ts_in = np.unique(atlas_in)[1:, np.newaxis]
    ts_in = np.repeat(ts_in, 2, axis=-1)

    assert (ts_in == ts_out).all()
    assert (atlas_in == atlas_out).all()
    assert (img_in.header["dim"] == img_out.header["dim"]).all()

    remove(atlastime)
    remove(atlas)


@mark.parametrize("ext", [(".nii.gz"), (".csv")])
def test_export_metrics_txt(ext, sc_mtx, atlas, sdi, testdir):
    testdir = join(testdir, "testdir")
    makedirs(testdir, exist_ok=True)
    atlas_in, mask, img = load_nifti_get_mask(atlas, ndim=3)
    mtx = read_mat(sc_mtx)
    mtx = mtx["SC_avg56"]
    sdi_in = np.genfromtxt(sdi)
    ts = np.unique(atlas_in)[1:]
    ts = np.repeat(ts[..., np.newaxis], 2, axis=-1)
    gsdi = dict.fromkeys(["hilo", "himid"], sdi_in)
    scgraph = SCGraph(mtx, ts, atlas=atlas_in, img=img, sdi=sdi_in)
    blocks.export_metric(scgraph, ext, join(testdir, "molly_"))
    scgraph = SCGraph(mtx, ts, atlas=atlas_in, img=img, gsdi=gsdi)
    blocks.export_metric(scgraph, ext, join(testdir, "cesar_"))

    assert isfile(join(testdir, f"molly_sdi{ext}"))
    assert isfile(join(testdir, f"cesar_gsdi_hilo{ext}"))
    assert isfile(join(testdir, f"cesar_gsdi_himid{ext}"))

    shutil.rmtree(testdir)
    remove(sc_mtx)
    remove(atlas)
    remove(sdi)


def test_export_metrics_nifti(sc_mtx, atlas, sdi, testdir):
    testdir = join(testdir, "testdir")
    makedirs(testdir, exist_ok=True)
    atlas_in, mask, img = load_nifti_get_mask(atlas, ndim=3)
    mtx = read_mat(sc_mtx)
    mtx = mtx["SC_avg56"]
    sdi_in = np.genfromtxt(sdi)
    ts = np.unique(atlas_in)[1:]
    ts = np.repeat(ts[..., np.newaxis], 2, axis=-1)
    scgraph = SCGraph(mtx, ts, atlas=atlas_in, sdi=sdi_in)
    blocks.export_metric(scgraph, ".nii.gz", join(testdir, "molly_"))

    assert isfile(join(testdir, "molly_sdi.tsv.gz"))

    sys.modules["nibabel"] = None
    scgraph = SCGraph(mtx, ts, atlas=atlas_in, img=img, sdi=sdi_in)
    blocks.export_metric(scgraph, ".nii.gz", join(testdir, "molly_"))
    sys.modules["nibabel"] = nibabel

    assert isfile(join(testdir, "molly_sdi.tsv.gz"))
    shutil.rmtree(testdir)
    remove(sc_mtx)
    remove(atlas)
    remove(sdi)


def test_plot_metrics(atlas, sc_mtx, sdi, testdir):
    testdir = join(testdir, "testdir")
    makedirs(testdir, exist_ok=True)
    atlas_in, _, img = load_nifti_get_mask(atlas, ndim=3)
    mtx = read_mat(sc_mtx)
    mtx = mtx["SC_avg56"]
    sdi_in = np.genfromtxt(sdi)
    ts = np.unique(atlas_in)[1:]
    ts = np.repeat(ts[..., np.newaxis], 2, axis=-1)
    gsdi = dict.fromkeys(["hilo", "himid"], sdi_in)

    scgraph = SCGraph(mtx, ts, sdi=sdi_in, img=img)
    blocks.plot_metric(scgraph, join(testdir, "maite_"), atlas=img)
    assert isfile(join(testdir, join(testdir, "maite_sdi.png")))

    atlas_in = find_parcellation_cut_coords(img)
    scgraph = SCGraph(mtx, ts, sdi=sdi_in, img=img)
    blocks.plot_metric(scgraph, join(testdir, "maite_"), atlas=atlas_in)
    assert isfile(join(testdir, join(testdir, "maite_sdi.png")))

    scgraph = SCGraph(mtx, ts, gsdi=gsdi, img=img)
    blocks.plot_metric(scgraph, join(testdir, "dimitri_"), atlas=img)
    assert isfile(join(testdir, join(testdir, "dimitri_gsdi_hilo.png")))

    shutil.rmtree(testdir)
    remove(sc_mtx)
    remove(atlas)
    remove(sdi)



#!/usr/bin/env python3
"""Tests for utils."""

from numpy import ndarray
from numpy.random import rand
from pytest import mark, raises

from nigsp import utils


# ### Unit tests
@mark.parametrize(
    "var, dtype",
    [
        (6, int),
        (4.2, float),
        ("hello", str),
        (rand(3), ndarray),
        ([1, 1, 2, 3, 5], list),
        ("6", int),
        (4, float),
        (42, str),
        ("hi", list),
        ([1, 1, 2, 3, 5], ndarray),
    ],
)
def test_change_var_type(var, dtype):
    """Test change_var_type."""
    # #!# Test logger!
    var_out = utils.change_var_type(var, dtype, stop=False)

    assert type(var_out) == dtype


# ### Break tests
@mark.parametrize(
    "var, dtype",
    [("6", int), (4, float), (42, str), ("hi", list), ([1, 1, 2, 3, 5], ndarray)],
)
def test_break_change_var_type(var, dtype):
    """Break change_var_type."""
    with raises(TypeError) as errorinfo:
        utils.change_var_type(var, dtype)
    assert "is not of type" in str(errorinfo.value)


def test_break_change_var_type_dtype():
    """Break change_var_type."""
    with raises(NotImplementedError) as errorinfo:
        utils.change_var_type(6, bool, stop=False)
    assert "not supported" in str(errorinfo.value)



"""
The core functions of `nigsp`.

They are organised into modules by type of operation.

Each function is callable through their respective module or directly
as an `operations` function, and each module can be called directly as a
`nigsp` module.

For example, calling `nigsp.operations.metrics.sdi` is equivalent to calling
`nigsp.operations.sdi` or `nigsp.metrisc.sdi`.
"""


# Import all operations.
from .graph import nodestrength, zerocross
from .laplacian import decomposition, symmetric_normalised_laplacian
from .metrics import functional_connectivity, gsdi, sdi
from .nifti import apply_atlas, apply_mask, mat_to_vol, unfold_atlas, unmask, vol_to_mat
from .surrogates import random_sign, sc_informed, sc_uninformed, test_significance
from .timeseries import (
    graph_filter,
    graph_fourier_transform,
    median_cutoff_frequency_idx,
    normalise_ts,
)



# -*- coding: utf-8 -*-
"""Parser for crispy-octo-broccoli."""

import argparse

from nigsp import __version__


def _get_parser():
    """
    Parse command line inputs for this function.

    Returns
    -------
    parser.parse_args() : argparse dict

    """
    parser = argparse.ArgumentParser(
        description=(
            "NiGSP, a tool to apply "
            "Graph Signal Processing to "
            "MRI (functional and structural), "
            "and compute derivative metrics "
            "such as the Structural Decoupling "
            "Index and the coupled/decoupled FC.\n"
            f"Version {__version__}"
        ),
        add_help=False,
    )
    required = parser.add_argument_group("Required Arguments")
    required.add_argument(
        "-f",
        "--input-func",
        dest="fname",
        type=str,
        help=(
            "Complete path (absolute or relative) and name "
            "of the file containing fMRI signal. This file "
            "can be a nifti, 1D, txt, csv, or mat file. "
            "If a nifti file is selected, an atlas has to "
            "be specified."
        ),
        required=True,
    )
    required.add_argument(
        "-s",
        "--input-structural",
        dest="scname",
        type=str,
        help=(
            "Complete path (absolute or relative) and name "
            "of the file containing the structural "
            "connectivity matrix. This file "
            "can be a 1D, txt, csv, or mat file."
        ),
        required=True,
    )

    opt_metrics = parser.add_argument_group(
        "Optional Arguments for metrics computation",
        description=(
            "Use these flag to "
            "select which metric "
            "should be computed. "
            "Note that the default "
            "behaviour is to compute "
            "all metrics."
        ),
    )
    opt_metrics.add_argument(
        "-sdi",
        "--structural-decoupling-index",
        dest="comp_metric",
        action="append_const",
        const="sdi",
        help=(
            "Compute the structural decoupling index "
            "(see Preti et al, 2019, Nat. Commun.)"
        ),
        default=None,
    )
    opt_metrics.add_argument(
        "-dfc",
        "--decoupled-functional-connectivity",
        dest="comp_metric",
        action="append_const",
        const="dfc",
        help=(
            "Compute the decoupledd functional connectivity "
            "(see Griffa et al, 2022, NeuroImage)"
        ),
        default=None,
    )

    opt_proc = parser.add_argument_group("Optional Arguments for data processing")
    opt_proc.add_argument(
        "-a",
        "--input-atlas",
        dest="atlasname",
        type=str,
        help=(
            "Complete path (absolute or relative) and name "
            "of the file containing the desired atlas. This "
            "file MUST be a nifti file, and it is required "
            "if the functional input is a nifti file. "
            "It is also required if the desired output is "
            "nifti. Default is None."
        ),
        default=None,
    )
    opt_proc.add_argument(
        "-idx",
        "--split-index",
        dest="index",
        type=str,
        help=(
            "The index of the harmonic that should split the "
            "harmonics in two, or the method to estimate such "
            'index. Default is "median", that invokes the '
            "estimation of the split index based on the energy "
            "of the graph."
        ),
        default="median",
    )

    opt_out = parser.add_argument_group("Optional Arguments for output")
    opt_out.add_argument(
        "-odir",
        "--output-directory",
        dest="outdir",
        type=str,
        help=(
            "Complete path (absolute or relative) and name "
            "of the desired output directory. If it does not "
            "exist, it will be created. If it is not "
            'specified, a folder named "%(prog)s" '
            "will be created in the folder containing the "
            "functional file."
        ),
        default=None,
    )
    opt_out.add_argument(
        "-ofile",
        "--output-file",
        dest="outname",
        type=str,
        help=(
            "Name of, and if desired path to, the output files."
            "Use this option to specify the file format of "
            "%(prog)s, and "
        ),
        default=None,
    )

    opt_stat = parser.add_argument_group("Optional Arguments for statistical test")

    opt_surr_type = opt_stat.add_mutually_exclusive_group()
    opt_surr_type.add_argument(
        "-sci",
        "--informed-surrogates",
        dest="surr_type",
        action="store_const",
        const="informed",
        help=(
            "Create surrogates informed of the structural "
            "connectivity matrix. Unless this option or "
            "the next one is selected, statistical test "
            "will be skipped"
        ),
        default=None,
    )
    opt_surr_type.add_argument(
        "-scu",
        "--uninformed-surrogates",
        dest="surr_type",
        action="store_const",
        const="uninformed",
        help=(
            "Create surrogates ignoring the structural "
            "connectivity matrix. Unless this option or "
            "the previous one is selected, statistical test "
            "will be skipped"
        ),
        default=None,
    )
    opt_stat.add_argument(
        "-n",
        "--surrogates-number",
        dest="n_surr",
        type=int,
        help=("Number of surrogates to create. Default is 1000."),
        default=1000,
    )
    opt_stat.add_argument(
        "-seed",
        "--random-seed",
        dest="seed",
        type=int,
        help=(
            "The seed to initialise. If None, internal "
            "seeds will be used depending on workflow, "
            "granting replicability. Default is None."
        ),
        default=None,
    )
    opt_stat.add_argument(
        "-method",
        "--statistical-method",
        dest="method",
        type=str,
        help=(
            "Type of statistical test to adopt. Possible "
            'options are "Bernoulli" (for a Bernoulli process '
            'test, group level only) or "frequentist" (for a '
            "frequentist approach, both group and subject level). "
            "Default is Bernoulli."
        ),
        default="Bernoulli",
    )

    optional = parser.add_argument_group("Other Optional Arguments")

    opt_logl = optional.add_mutually_exclusive_group()
    opt_logl.add_argument(
        "-debug",
        "--debug",
        dest="lgr_degree",
        action="store_const",
        const="debug",
        help="Only print debugging info to log file.",
        default="info",
    )
    opt_logl.add_argument(
        "-quiet",
        "--quiet",
        dest="lgr_degree",
        action="store_const",
        const="quiet",
        help="Only print warnings to log file.",
        default="info",
    )
    optional.add_argument(
        "-h", "--help", action="help", help="Show this help message and exit"
    )
    optional.add_argument(
        "-v", "--version", action="version", version=("%(prog)s " + __version__)
    )
    return parser


if __name__ == "__main__":
    raise RuntimeError(
        "nigsp/cli/run.py should not be run directly;\n"
        "Please `pip install` nigsp and use the "
        "`nigsp` command"
    )


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Operations on graphs and graph derived properties.

Attributes
----------
LGR
    Logger
"""

import logging
from itertools import combinations

import numpy as np

from nigsp import references
from nigsp.due import due
from nigsp.utils import prepare_ndim_iteration

LGR = logging.getLogger(__name__)

SUPPORTED_METRICS = ["sdi", "dfc", "fc"]


@due.dcite(references.PRETI_2019)
def sdi(ts_split, mean=False, keys=None):
    """
    Compute the Structural Decoupling Index (SDI).

    i.e. the ratio between the norms of the "high" and the norm of the "low"
    "graph-filtered" timeseries.

    If the given dictionary does not contain the keywords "high" and "low",
    the SDI is computed as the ratio between the norm of the second and
    the norm of the first dictionary entry.
    "keys" can be used to indicate the order of the two keys, or to select two
    elements of a bigger dictionary.

    Parameters
    ----------
    ts_split : dict or numpy.ndarrays
        A dictionary containing two entries. If the two entries are "low" and
        "high", then SDI will be computed as the norm of the high vs the norm
        of the low, otherwise as the ratio between the second (second key in
        sorted keys) and the first.
    mean : bool, optional
        If True, compute mean over the last axis (e.g. between subjects)
    keys : None or list of strings, optional
        Can be used to select two entries from a bigger dictionary
        and/or to specify the order in which the keys should be read (e.g.
        forcing a different order from the sorted keys).

    Returns
    -------
    numpy.ndarray
        Returns the structural decoupling index

    Raises
    ------
    ValueError
        If keys are provided but not contained in the dictionary
        If keys are not provided and the dictionary has more than 2 entries
    """
    # #!# Implement acceptance of two matrices and not only dictionary
    if keys is None:
        keys = list(ts_split.keys())
    else:
        if all(item in list(ts_split.keys()) for item in keys) is False:
            raise ValueError(
                f"The provided keys {keys} do not match the "
                "keys of the provided dictionary "
                f"({list(ts_split.keys())})"
            )

    if len(keys) != 2:
        raise ValueError(
            "`structural_decoupling_index` function requires "
            "a dictionary with exactly two timeseries as input."
        )

    check_keys = [item.lower() for item in keys]
    if all(item in ["low", "high"] for item in check_keys):
        # Case insensitively reorder the items of dictionary as ['low', 'high'].
        keys = [keys[check_keys.index("low")], keys[check_keys.index("high")]]

    norm = dict.fromkeys(keys)
    for k in keys:
        norm[k] = np.linalg.norm(ts_split[k], axis=1)

    LGR.info("Computing Structural Decoupling Index.")
    sdi = norm[keys[1]] / norm[keys[0]]

    if sdi.ndim >= 2 and mean:
        sdi = sdi.mean(axis=1)

    return np.log2(sdi)


def gsdi(ts_split, mean=False, keys=None):
    """Compute the generalised SDI.

    Parameters
    ----------
    ts_split : dict or numpy.ndarrays
        A dictionary containing two entries. If the two entries are "low" and
        "high", then SDI will be computed as the norm of the high vs the norm
        of the low, otherwise as the ratio between the second (second key in
        sorted keys) and the first.
    mean : bool, optional
        If True, compute mean over the last axis (e.g. between subjects)
    keys : None or list of strings, optional
        Can be used to select two entries from a bigger dictionary
        and/or to specify the order in which the keys should be read (e.g.
        forcing a different order from the sorted keys).

    Returns
    -------
    dict of numpy.ndarray
        Returns a dictionary of computed gSDI.

    Raises
    ------
    ValueError
        If keys are provided but not contained in the dictionary.
    """
    # #!# Implement acceptance of N matrices and not only dictionary
    # #!# Check that this can overcome SDI
    if keys is None:
        keys = list(ts_split.keys())
    else:
        if all(item in list(ts_split.keys()) for item in keys) is False:
            raise ValueError(
                f"The provided keys {keys} do not match the "
                "keys of the provided dictionary "
                f"({list(ts_split.keys())})"
            )

    if len(keys) > 2:
        LGR.info("Prepare combinations of timeseries")
        # Combining two or more "timeseries splits" means nothing more than
        # adding them in this case.
        for n in range(2, len(keys)):
            for c in combinations(keys, n):
                comb_key = str(c).replace("'", "").replace(", ", "_and_")

                ts_split[comb_key] = np.zeros_like(ts_split[keys[0]], dtype="float32")
                for k in c:
                    ts_split[comb_key] = np.add(ts_split[comb_key], ts_split[k])

    # Obtain updated list of keys
    all_keys = list(ts_split.keys())
    norm = dict.fromkeys(all_keys)
    LGR.info("Computing norm of timeseries")
    for k in all_keys:
        norm[k] = np.linalg.norm(ts_split[k], axis=1)

    LGR.info("Computing generalised SDI")
    gsdi = dict()
    for k in keys:
        for j in all_keys:
            if not (k in j or f"({k}_" in j or f"_{k})" in j or f"_{k}_" in j):
                gsdi[f"{k}_over_{j}"] = norm[k] / norm[j]

    if list(gsdi.values())[0].ndim >= 2 and mean:
        for k in gsdi.keys():
            gsdi[k] = np.log2(gsdi[k].mean(axis=1))
    else:
        for k in gsdi.keys():
            gsdi[k] = np.log2(gsdi[k])

    return gsdi


@due.dcite(references.GRIFFA_2022)
def functional_connectivity(timeseries, mean=False):
    """
    Compute Functional Connectivity of timeseries.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray or dict of numpy.ndarray
        Adictionary of (or a single) 2- or 3-D matrix with timeseries along axis 1.
    mean : bool, optional
        If timeseries is 3D and this is True, return the average FC along the last axis.

    Returns
    -------
    numpy.ndarray or dict of numpy.ndarray
        Functional Connectivity of the given timeseries input.
    """

    def _fc(timeseries, mean=False):
        """
        Quick functional connectivity computation.

        This is not meant to be used outside of other functions.

        It is assumed that time is encoded in the second dimension (axis 1),
        e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

        Parameters
        ----------
        timeseries : numpy.ndarray
            A 2- or 3-D matrix containing timeseries along axis 1.
        mean : bool, optional
            If timeseries is 3D and this is True, return the average FC along the last axis.

        Returns
        -------
        numpy.ndarray
            FC matrix
        """
        timeseries = timeseries.squeeze()
        if timeseries.ndim < 2:
            LGR.warning("Computing functional connectivity of a 1D array (== 1)!")
            return 1
        elif timeseries.ndim == 2:
            return np.corrcoef(timeseries)
        else:
            # reshape the array to allow reiteration on unknown dimensions of timeseries
            temp_ts, _ = prepare_ndim_iteration(timeseries, 2)
            fcorr = np.empty(
                ([temp_ts.shape[0]] * 2 + [temp_ts.shape[-1]]), dtype="float32"
            )
            for i in range(temp_ts.shape[-1]):
                fcorr[:, :, i] = np.corrcoef(temp_ts[..., i])

            if timeseries.ndim > 3:
                new_shape = (timeseries.shape[0],) * 2 + timeseries.shape[2:]
                fcorr = fcorr.reshape(new_shape)

            if mean:
                fcorr = fcorr.mean(axis=2).squeeze()
            return fcorr

    if type(timeseries) is dict:
        fc = dict()
        for k in timeseries.keys():
            fc[k] = _fc(timeseries[k], mean)
    else:
        fc = _fc(timeseries, mean)

    return fc


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for operations.metrics."""

import numpy as np
from numpy.random import rand
from pytest import raises

from nigsp.operations import metrics
from nigsp.utils import prepare_ndim_iteration


# ### Unit tests
def test_sdi():
    ts1 = np.arange(1, 3)[..., np.newaxis]
    ts2 = np.arange(3, 5)[..., np.newaxis]
    ts3 = np.arange(5, 7)[..., np.newaxis]
    sdi_in = np.log2(np.arange(3.0, 1.0, -1.0))

    ts = {"low": ts1, "high": ts2}
    sdi_out = metrics.sdi(ts)
    assert (sdi_out == sdi_in).all()

    ts = {"HIGH": ts2, "LOW": ts1}
    sdi_out = metrics.sdi(ts)
    assert (sdi_out == sdi_in).all()

    ts = {"alpha": ts1, "beta": ts2, "gamma": ts3}
    sdi_out = metrics.sdi(ts, keys=["alpha", "beta"])
    assert (sdi_out == sdi_in).all()

    ts = {
        "low": np.repeat(np.repeat(ts1[..., np.newaxis], 3, axis=1), 3, axis=2),
        "high": np.repeat(np.repeat(ts2[..., np.newaxis], 3, axis=1), 3, axis=2),
    }
    sdi_out = metrics.sdi(ts, mean=True)
    sdi_out = np.around(sdi_out, decimals=15)
    assert (sdi_out == sdi_in).all()


def test_gsdi():
    ts1 = np.arange(1, 3)[..., np.newaxis]
    ts2 = np.arange(3, 5)[..., np.newaxis]
    ts3 = np.arange(5, 7)[..., np.newaxis]
    ts = {"alpha": ts1, "beta": ts2, "gamma": ts3}
    gsdi_in = np.log2(np.arange(3.0, 1.0, -1.0))
    gsdi_and_in = np.log2(
        np.linalg.norm(ts1, axis=1) / np.linalg.norm(np.add(ts2, ts3), axis=1)
    )
    # fmt: off
    keys_in = ['alpha_over_beta', 'alpha_over_gamma', 'alpha_over_(beta_and_gamma)',
               'beta_over_alpha', 'beta_over_gamma', 'beta_over_(alpha_and_gamma)',
               'gamma_over_alpha', 'gamma_over_beta', 'gamma_over_(alpha_and_beta)']
    # fmt: on
    gsdi_out = metrics.gsdi(ts)

    assert all(item in keys_in for item in list(gsdi_out.keys()))
    assert (gsdi_out["beta_over_alpha"] == gsdi_in).all()
    assert (gsdi_out["alpha_over_(beta_and_gamma)"] == gsdi_and_in).all()

    gsdi_out = metrics.gsdi(ts, keys=["alpha", "beta"])
    assert all(
        item in list(gsdi_out.keys()) for item in ["alpha_over_beta", "beta_over_alpha"]
    )
    assert (gsdi_out["beta_over_alpha"] == gsdi_in).all()

    ts = {
        "alpha": np.repeat(np.repeat(ts1[..., np.newaxis], 3, axis=1), 3, axis=2),
        "beta": np.repeat(np.repeat(ts2[..., np.newaxis], 3, axis=1), 3, axis=2),
    }
    gsdi_out = metrics.gsdi(ts, mean=True)
    gsdi_out["beta_over_alpha"] = np.around(gsdi_out["beta_over_alpha"], decimals=15)
    assert (gsdi_out["beta_over_alpha"] == gsdi_in).all()


# ### Break tests
def test_break_sdi():
    ts1 = np.arange(1, 3)[..., np.newaxis]
    ts2 = np.arange(3, 5)[..., np.newaxis]
    ts3 = np.arange(5, 7)[..., np.newaxis]
    ts = {"alpha": ts1, "beta": ts2, "gamma": ts3}

    with raises(ValueError) as errorinfo:
        metrics.sdi(ts, keys=["high", "low"])
    assert "provided keys" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        metrics.sdi(ts)
    assert "exactly two" in str(errorinfo.value)


def test_break_gsdi():
    ts1 = np.arange(1, 3)[..., np.newaxis]
    ts2 = np.arange(3, 5)[..., np.newaxis]
    ts3 = np.arange(5, 7)[..., np.newaxis]
    ts = {"alpha": ts1, "beta": ts2, "gamma": ts3}

    with raises(ValueError) as errorinfo:
        metrics.gsdi(ts, keys=["physio", "lambda", "pi"])
    assert "provided keys" in str(errorinfo.value)


def test_functional_connectivity():
    assert metrics.functional_connectivity(rand(6)) == 1

    ts = rand(3, 6, 2, 2)

    tst, _ = prepare_ndim_iteration(ts, 2)
    fc = np.empty(([tst.shape[0]] * 2 + [tst.shape[-1]]), dtype="float32")
    for i in range(tst.shape[-1]):
        fc[:, :, i] = np.corrcoef(tst[..., i])

    ns = (ts.shape[0],) * 2 + ts.shape[2:]
    fc = fc.reshape(ns).mean(axis=2).squeeze()

    fc_t = metrics.functional_connectivity(ts, mean=True)

    assert (fc == fc_t).all()

    tsd = {"hi": rand(3, 6), "lo": rand(3, 6)}

    fcd = metrics.functional_connectivity(tsd)

    assert all(item in list(tsd.keys()) for item in list(fcd.keys()))

    for k in fcd.keys():
        assert (fcd[k] == np.corrcoef(tsd[k])).all()



#!/usr/bin/env python3
"""
Operations for timeseries.

Attributes
----------
LGR
    Logger
"""

import logging

import numpy as np

from nigsp.utils import change_var_type, pairwise, prepare_ndim_iteration

LGR = logging.getLogger(__name__)


def normalise_ts(timeseries, globally=False):
    """
    Normalise given timeseries (i.e. mean=0, std=1).

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Any timeseries with std == 0 is returned as a series of 0s.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    globally : bool, optional
        If True, normalise timeseries across the first two axes

    Returns
    -------
    numpy.ndarray
        The normalised timeseries (mean=0 std=1) if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.
    """
    if timeseries.ndim < 2 or (timeseries.ndim == 2 and timeseries.shape[1] == 1):
        LGR.warning(
            "Given timeseries seems to be a single timepoint. Returning it as is."
        )
        return timeseries

    if globally:
        z = (timeseries - timeseries.mean(axis=(0, 1))) / timeseries.std(
            axis=(0, 1), ddof=1
        )
    else:
        z = (timeseries - timeseries.mean(axis=1)[:, np.newaxis, ...]) / timeseries.std(
            axis=1, ddof=1
        )[:, np.newaxis, ...]

    z[np.isnan(z)] = 0

    return z


def spc_ts(timeseries, globally=False):
    """
    Express timeseries in signal percentage change.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    globally : bool, optional
        If True, SPC timeseries across the first two axes (mainly for similarity with other functions.)

    Returns
    -------
    numpy.ndarray
        The timeseries in SPC if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.
    """
    if timeseries.ndim < 2 or (timeseries.ndim == 2 and timeseries.shape[1] == 1):
        LGR.warning(
            "Given timeseries seems to be a single timepoint. Returning it as is."
        )
        return timeseries

    if globally:
        scp = (timeseries - timeseries.mean(axis=(0, 1))) / timeseries.mean(axis=(0, 1))
    else:
        scp = (
            timeseries - timeseries.mean(axis=1)[:, np.newaxis, ...]
        ) / timeseries.mean(axis=1)[:, np.newaxis, ...]

    scp[np.isnan(scp)] = timeseries[np.isnan(scp)]

    return scp


def demean_ts(timeseries, globally=False):
    """
    Demean timeseries.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    globally : bool, optional
        If True, demean timeseries across the first two axes.

    Returns
    -------
    numpy.ndarray
        The demeaned timeseries if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.
    """
    if timeseries.ndim < 2 or (timeseries.ndim == 2 and timeseries.shape[1] == 1):
        LGR.warning(
            "Given timeseries seems to be a single timepoint. Returning it as is."
        )
        return timeseries

    if globally:
        return timeseries - timeseries.mean(axis=(0, 1))
    else:
        return timeseries - timeseries.mean(axis=1)[:, np.newaxis, ...]


def rescale_ts(timeseries, vmin=0, vmax=1, globally=False):
    """
    Rescale given timeseries between given max and min value.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Any timeseries with std == 0 is returned as a series of 0s.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    vmin : float, optional
        The minimum value to scale between
    vmax : float, optional
        The maximum value to scale between
    globally : bool, optional
        If True, rescale timeseries across the first two axes

    Returns
    -------
    numpy.ndarray
        The normalised timeseries (mean=0 std=1) if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.
    """
    if timeseries.ndim < 2 or (timeseries.ndim == 2 and timeseries.shape[1] == 1):
        LGR.warning(
            "Given timeseries seems to be a single timepoint. Returning it as is."
        )
        return timeseries

    if globally:
        res = timeseries - timeseries.min(axis=(0, 1))
        res = res / res.max(axis=(0, 1))
    else:
        res = timeseries - timeseries.min(axis=1)[:, np.newaxis, ...]
        res = res / res.max(axis=1)[:, np.newaxis, ...]

    res = res * (vmax - vmin) + vmin
    return res


def resize_ts(timeseries, resize=None, globally=False):
    """
    Rescale timeseries with some methods.

    It is assumed that time is encoded in the second dimension (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    resize : 'spc', 'norm', 'gnorm', 'demean', 'gdemean' tuple, list, or None, optional
        Whether to resize the signal or not before plotting.
        If 'spc', compute signal percentage change
        If 'norm', normalise signals (z-score)
        If 'demean', remove signal average
        If 'gsr', remove global signal (average across points)
        If tuple or list, rescale signals between those two values
        If None, don't do anything (default)
    globally : bool, optional
        If True, rescale timeseries across the first two axes

    Returns
    -------
    numpy.ndarray
        The timeseries after resizing if timeseries is not a 1D array.
        If timeseries is a 1D array, it is returned as is.

    """
    if resize and timeseries.ndim > 1:
        if resize == "spc":  # pragma: no cover
            LGR.info("Expressing timeseries in signal percentage change")
            timeseries = spc_ts(timeseries, globally=globally)
        elif resize == "norm":  # pragma: no cover
            LGR.info("Normalise timeseries")
            timeseries = normalise_ts(timeseries, globally=globally)
        elif resize == "demean":  # pragma: no cover
            LGR.info("Demean timeseries")
            timeseries = demean_ts(timeseries, globally=globally)
        elif resize == "gsr":
            LGR.info("Remove Global Signal from timeseries")
            timeseries = timeseries - timeseries.mean(axis=0)
        elif type(resize) in [tuple, list]:
            if len(resize) != 2:
                raise NotImplementedError("Required two elements to express rescaling")
            LGR.info(
                f"Expressing timeseries in given range {resize}"
            )  # pragma: no cover
            timeseries = resize(
                timeseries, vmin=resize[0], vmax=resize[1], globally=globally
            )  # pragma: no cover
        else:
            raise NotImplementedError(
                f"Chosen rescaling method {resize} is not supported."
            )

    return timeseries


def graph_fourier_transform(timeseries, eigenvec, energy=False, mean=False):
    """
    Project a graph decomposition onto the timeseries.

    It returns the result of the projection or the energy of the spectral
    density of the projection.

    If `mean` is true and the timeseries has 3 dimensions, it returns the
    mean across the last dimension.

    It is assumed that time is encoded in the second dimension of timeseries (axis 1),
    e.g. for 90 voxels and 300 timepoints, shape is [90, 300].

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension (axis 1) is time.
    eigenvec : numpy.ndarray
        The eigenvector resulting from the Laplacian decomposition.
    energy : bool, optional
        If True, returns the energy (power) of the spectral density instead of
        the projection.
    mean : bool, optional
        If True and timeseries has 3 dimensions, returns the mean across axis 1 (time).

    Returns
    -------
    np.ndarray
        Returns either the projection of the graph on the timeseries, or its energy.
    """
    timeseries = timeseries.squeeze()
    if timeseries.ndim < 3:
        proj = eigenvec.conj().T @ timeseries
    else:
        temp_ts, proj = prepare_ndim_iteration(timeseries, 2)
        for i in range(temp_ts.shape[-1]):
            proj[:, :, i] = eigenvec.conj().T @ np.squeeze(temp_ts[:, :, i])
        if timeseries.ndim > 3:
            proj = proj.reshape(timeseries.shape)

    if timeseries.ndim > 2 and mean and not energy:
        proj = proj.mean(axis=1)

    if energy:
        # Compute energy of the spectral density
        energy = proj**2
        if proj.ndim > 2 and mean:
            energy = energy.mean(axis=1)

        return energy

    else:
        return proj


def median_cutoff_frequency_idx(energy):
    """
    Find the frequency that splits the energy of a timeseries in two roughly equal parts.

    Parameters
    ----------
    energy : numpy.ndarray
        The array representing the energy (power) spectral density of a timeseries.
        this array can be 1D or 2D - if 2D it's assumed that the second dimension
        represents subjects.

    Returns
    -------
    int
        The index of the frequency that splits the spectral power into two
        (more or less) equal parts.

    Raises
    ------
    NotImplementedError
        If the provided array is 3D or more.
    """
    if energy.ndim > 2:
        raise NotImplementedError(
            "Provided energy spectral density data have "
            f"{energy.ndim} dimensions, but arrays of more "
            "than 2 dimensions are not supported yet"
        )

    if energy.ndim == 2:
        energy = energy.mean(axis=-1)
    half_tot_auc = np.trapz(energy, axis=0) / 2
    LGR.debug(f"Total AUC = {half_tot_auc*2}, targeting half of total AUC")

    # Compute the AUC from first to one to last frequency,
    # skipping first component because AUC(1)=0.
    # Returns first idx for which AUC reaches half of total AUC.
    # It could be computed from high to low, but in theory
    # there would be more cycles, as #eigenvec(h) > #eigenvec(l).
    for freq_idx in range(1, energy.size):
        LGR.debug(
            f"Frequency idx {freq_idx}, "
            f"AUC = {np.trapz(energy[:freq_idx])}, "
            f"target AUC = {half_tot_auc}"
        )
        if np.trapz(energy[:freq_idx]) >= half_tot_auc:
            break

    LGR.info(f"Found {freq_idx} as splitting index")
    return freq_idx


def graph_filter(timeseries, eigenvec, freq_idx, keys=["low", "high"]):
    """
    Filter a graph decomposition into two parts based on freq_idx.

    Return the two eigenvector lists (high freq and low freq) that are equal
    to the original eigenvector list, but "low" is zero-ed for all frequencies
    >= of the given index, and "high" is zero-ed for all frequencies < to the
    given index.
    Also return their projection onto a timeseries.

    Parameters
    ----------
    timeseries : numpy.ndarray
        The input timeseries. It is assumed that the second dimension is time.
    eigenvec : numpy.ndarray
        The eigenvector resulting from the Laplacian decomposition.
    freq_idx : int or list
        The index of the frequency that splits the spectral power into two
        (more or less) equal parts - i.e. the index of the first frequency in
        the "high" component.
    keys : list, optional
        The keys to call the split parts with

    Returns
    -------
    dict of numpy.ndarray
        Return first the split eigenvectors
    dict of numpy.ndarray
        Return second the projected split eigenvectors onto the timeseries.

    Raises
    ------
    IndexError
        If the given index is 0 (all "high"), the last possible index (all "low"),
        or higher than the last possible index (not applicable).
    """
    # #!# Find better name
    # #!# Implement an index splitter
    freq_idx = change_var_type(freq_idx, list, stop=False, silent=True)

    for f in freq_idx:
        if f == 0 or f >= eigenvec.shape[0] - 1:
            raise IndexError(
                f"Selected index {f} is not valid to split "
                f"eigenvector matrix of shape {eigenvec.shape}."
            )

    LGR.info(f"Splitting graph into {len(freq_idx)+1} parts")

    # Check that there is the right amount of keys
    if len(keys) > len(freq_idx) + 1:
        LGR.warning(
            f"The declared keys list ({keys}) has {len(keys)} elements. "
            f"Since the frequency index list ({freq_idx}) has {len(freq_idx)}, "
            f"any keys after {keys[len(freq_idx)]} will be ignored."
        )
        keys = keys[: len(freq_idx) + 1]
    elif len(keys) < len(freq_idx) + 1:
        LGR.warning(
            f"The declared keys list ({keys}) has {len(keys)} elements. "
            f"Since the frequency index list ({freq_idx}) has {len(freq_idx)}, "
            f"more keys will be created after {keys[len(freq_idx)]} ."
        )

        for i in range(len(keys), len(freq_idx) + 1):
            keys = keys + [f"key-{i+1:03d}"]

    # Add 0 and None to freq_idx to have full indexes
    freq_idx = [0] + freq_idx + [None]

    evec_split = dict.fromkeys(keys)
    ts_split = dict.fromkeys(keys)

    for n, idx in enumerate(pairwise(freq_idx)):
        i, j = idx
        k = j if j is not None else eigenvec.shape[-1]
        evec_split[keys[n]] = np.append(
            np.append(
                np.zeros_like(eigenvec[:, :i], dtype="float32"),
                eigenvec[:, i:j],
                axis=-1,
            ),
            np.zeros_like(eigenvec[:, k:], dtype="float32"),
            axis=-1,
        )

    LGR.info("Compute graph fourier coefficients.")
    fourier_coeff = graph_fourier_transform(timeseries, eigenvec)

    for k in keys:
        LGR.info(f"Compute {k} part of timeseries.")
        ts_split[k] = graph_fourier_transform(fourier_coeff, evec_split[k].T)

    return evec_split, ts_split


"""
Copyright 2022, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for viz."""

import sys
from os import remove
from os.path import isfile

import matplotlib
import nilearn.plotting
from numpy import genfromtxt
from numpy.random import rand
from pytest import mark, raises

from nigsp import viz


@mark.parametrize(
    "mtx", [(rand(3, 4)), (rand(4, 4)), (rand(4, 4, 3)), (rand(4, 4, 3, 1))]
)
# ### Unit tests
def test_plot_connectivity(mtx):
    viz.plot_connectivity(mtx, "annie.png", closeplot=True)

    assert isfile("annie.png")
    matplotlib.pyplot.close()
    remove("annie.png")


@mark.parametrize("timeseries", [(rand(3, 50)), (rand(3, 50, 3)), (rand(3, 50, 4, 1))])
def test_plot_greyplot(timeseries):
    viz.plot_greyplot(timeseries, "troy.png", closeplot=True)

    assert isfile("troy.png")
    matplotlib.pyplot.close()
    remove("troy.png")


def test_plot_nodes(sdi, atlas):
    ns = genfromtxt(sdi)
    viz.plot_nodes(ns, atlas, "abed.png", closeplot=True)

    assert isfile("abed.png")
    matplotlib.pyplot.close()
    remove("abed.png")
    remove(sdi)
    remove(atlas)


def test_plot_edges(atlas):
    mtx = rand(360, 360)
    mtx = mtx - mtx.min() + 0.3
    viz.plot_edges(mtx, atlas, "britta.png", thr="95%", closeplot=True)

    assert isfile("britta.png")
    matplotlib.pyplot.close()
    remove("britta.png")
    remove(atlas)


# ### Break tests
def test_break_plot_connectivity():
    sys.modules["matplotlib"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_connectivity(rand(3, 3), "craig.png")
    assert "are required" in str(errorinfo.value)
    sys.modules["matplotlib"] = matplotlib

    sys.modules["nilearn.plotting"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_connectivity(rand(3, 3), "craig.png")
    assert "are required" in str(errorinfo.value)
    sys.modules["nilearn.plotting"] = nilearn.plotting

    with raises(ValueError) as errorinfo:
        viz.plot_connectivity(rand(3, 3, 3, 4), "craig.png")
    assert "plot connectivity" in str(errorinfo.value)

    matplotlib.pyplot.close()


def test_break_plot_greyplot():
    sys.modules["matplotlib"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_greyplot(rand(3, 3), "dan.png")
    assert "is required" in str(errorinfo.value)
    sys.modules["matplotlib"] = matplotlib

    with raises(ValueError) as errorinfo:
        viz.plot_greyplot(rand(3, 3, 3, 4), "dan.png")
    assert "plot greyplots" in str(errorinfo.value)

    matplotlib.pyplot.close()


def test_break_plot_nodes(atlas):
    sys.modules["matplotlib"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_nodes(rand(3), rand(3, 3))
    assert "are required" in str(errorinfo.value)
    sys.modules["matplotlib"] = matplotlib

    sys.modules["nilearn.plotting"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_nodes(rand(3), rand(3, 3))
    assert "are required" in str(errorinfo.value)
    sys.modules["nilearn.plotting"] = nilearn.plotting

    with raises(ValueError) as errorinfo:
        viz.plot_nodes(rand(3, 3, 4), rand(3, 3))
    assert "plot node values" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        viz.plot_nodes(rand(3), rand(3, 3, 2))
    assert "atlases in nifti" in str(errorinfo.value)
    with raises(NotImplementedError) as errorinfo:
        viz.plot_nodes(rand(3), rand(3, 4))
    assert "atlases in nifti" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        viz.plot_nodes(rand(3), rand(4, 3))
    assert "different length" in str(errorinfo.value)

    matplotlib.pyplot.close()
    remove(atlas)


def test_break_plot_edges(atlas):
    sys.modules["matplotlib"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_edges(rand(3), rand(3, 3))
    assert "are required" in str(errorinfo.value)
    sys.modules["matplotlib"] = matplotlib

    sys.modules["nilearn.plotting"] = None
    with raises(ImportError) as errorinfo:
        viz.plot_edges(rand(3), rand(3, 3))
    assert "are required" in str(errorinfo.value)
    sys.modules["nilearn.plotting"] = nilearn.plotting

    with raises(ValueError) as errorinfo:
        viz.plot_edges(rand(3, 3, 4, 5), rand(3, 3))
    assert "plot node values" in str(errorinfo.value)

    with raises(NotImplementedError) as errorinfo:
        viz.plot_edges(rand(3), rand(3, 3, 2))
    assert "atlases in nifti" in str(errorinfo.value)
    with raises(NotImplementedError) as errorinfo:
        viz.plot_edges(rand(3), rand(3, 4))
    assert "atlases in nifti" in str(errorinfo.value)

    with raises(ValueError) as errorinfo:
        viz.plot_edges(rand(3), rand(4, 3))
    assert "different length" in str(errorinfo.value)

    matplotlib.pyplot.close()
    remove(atlas)


# ..and a movie!!!
